# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
from os import environ

import numpy as np
from smartredis import Client, Dataset
from smartredis.error import RedisReplyError

from ..error import SSInternalError
from ..log import get_logger

logger = get_logger(__name__)


def form_name(*args):
    return "_".join(str(arg) for arg in args if arg is not None)


class DataInfo:
    """A class holding all relevant information to download datasets from aggregation lists

    This class can be passed as argument to SmartSim's ML data loaders, as it wraps the
    information about the aggregation list holding the training datasets.

    Each training dataset will store batches of samples and (optionally) labels or targets in the
    form of tensors. The tensors must always have the same names, which can be accessed
    in ``DataInfo.sample_name`` and ``DataInfo.target_name``.

    :param list_name: Name of the aggregation list used for sample datasets
    :type list_name: str
    :param sample_name: Name of tensor holding training samples in stored datasets.
    :type sample_name: str
    :param target_name: Name of tensor holding targets or labels in stored datasets.
    :type target_name: str
    :num_classes: Number of classes (for categorical data).
    :type num_classes: int | None
    """

    def __init__(
        self, list_name, sample_name="samples", target_name="targets", num_classes=None
    ):
        self.list_name = list_name
        self.sample_name = sample_name
        self.target_name = target_name
        self.num_classes = num_classes
        self._ds_name = form_name(self.list_name, "info")

    def publish(self, client: Client):
        """Upload DataInfo information to Orchestrator

        The information is put on the DB as a DataSet, with strings
        stored as metastrings and integers stored as metascalars.

        :param client: Client to connect to Database
        :type client: SmartRedis.Client
        """
        info_ds = Dataset(self._ds_name)
        info_ds.add_meta_string("sample_name", self.sample_name)
        if self.target_name:
            info_ds.add_meta_string("target_name", self.target_name)
        if self.num_classes:
            info_ds.add_meta_scalar("num_classes", self.num_classes)
        client.put_dataset(info_ds)

    def download(self, client: Client):
        """Download DataInfo information from Orchestrator

        The information retrieved from the DB is used to populate
        this object's members. If the information is not available
        on the DB, the object members are not modified.

        :param client: Client to connect to Database
        :type client: SmartRedis.Client
        """
        try:
            info_ds = client.get_dataset(self._ds_name)
        except RedisReplyError:
            # If the info was not published, proceed with default parameters
            logger.warning(
                f"Could not retrieve data for DataInfo object, the following values will be kept."
            )
            logger.warning(str(self))
            return
        self.sample_name = info_ds.get_meta_strings("sample_name")[0]
        field_names = info_ds.get_metadata_field_names()
        if "target_name" in field_names:
            self.target_name = info_ds.get_meta_strings("target_name")[0]
        if "num_classes" in field_names:
            self.num_classes = info_ds.get_meta_scalars("num_classes")[0]

    def __repr__(self):
        strings = ["DataInfo object"]
        strings += [f"Aggregation list name: {self.list_name}"]
        strings += [f"Sample tensor name: {self.sample_name}"]
        if self.target_name:
            strings += [f"Target tensor name: {self.target_name}"]
        if self.num_classes:
            strings += [f"Number of classes: {self.num_classes}"]
        return "\n".join(strings)


class TrainingDataUploader:
    """A class to simplify uploading batches of samples to train a model.

    This class can be used to upload samples following a simple convention
    for naming. Once created, the function `publish_info` can be used
    to put all details about the data set on the Orchestrator. A training
    process can thus access them and get all relevant information to download
    the batches which are uploaded.

    Each time a new batch is available, it is sufficient to call `put_batch`,
    and the data will be stored following the naming convention specified
    by the attributes of this class.

    :param list_name: Name of the dataset as stored on the Orchestrator
    :type list_name: str
    :param sample_name: Name of samples tensor in uploaded Datasets
    :type sample_name: str
    :param target_name: Name of targets tensor (if needed) in uploaded Datasets
    :type target_name: str
    :param num_classes: Number of classes of targets, if categorical
    :type num_classes: int
    :param cluster: Whether the SmartSim Orchestrator is being run as a cluster
    :type cluster: bool
    :param address: Address of Redis DB as <ip_address>:<port>
    :type address: str
    :param rank: Rank of DataUploader in multi-process application (e.g. MPI rank).
    :type rank: int
    :param verbose: If output should be logged to screen.
    :type verbose: bool

    """

    def __init__(
        self,
        list_name="training_data",
        sample_name="samples",
        target_name="targets",
        num_classes=None,
        cluster=True,
        address=None,
        rank=None,
        verbose=False,
    ):
        if not list_name:
            raise ValueError("Name can not be empty")
        if not sample_name:
            raise ValueError("Sample name can not be empty")

        self.client = Client(address=address, cluster=cluster)
        self.verbose = verbose
        self.batch_idx = 0
        self.rank = rank
        self._info = DataInfo(list_name, sample_name, target_name, num_classes)

    @property
    def list_name(self):
        return self._info.list_name

    @property
    def sample_name(self):
        return self._info.sample_name

    @property
    def target_name(self):
        return self._info.target_name

    @property
    def num_classes(self):
        return self._info.num_classes

    def publish_info(self):
        self._info.publish(self.client)

    def put_batch(self, samples, targets=None):
        batch_ds_name = form_name("training_samples", self.rank, self.batch_idx)
        batch_ds = Dataset(batch_ds_name)
        batch_ds.add_tensor(self.sample_name, samples)
        if (
            targets is not None
            and self.target_name
            and (self.target_name != self.sample_name)
        ):
            batch_ds.add_tensor(self.target_name, targets)
            if self.verbose:
                logger.info(f"Putting dataset {batch_ds_name} with samples and targets")
        else:
            if self.verbose:
                logger.info(f"Putting dataset {batch_ds_name} with samples")

        self.client.put_dataset(batch_ds)
        self.client.append_to_list(self.list_name, batch_ds)
        if self.verbose:
            logger.info(f"Added dataset to list {self.list_name}")
            logger.info(f"List length {self.client.get_list_length(self.list_name)}")

        self.batch_idx += 1


class DataDownloader:
    """A class to download a dataset from the DB.

    By default, the DataDownloader has to be created in a process
    launched through SmartSim, with sample producers listed as incoming
    entities.

    Information about the uploaded datasets can be defined in two ways:

     - By supplying a DataInfo object as value of ``data_info_or_list_name``

     - By supplying a string as value of ``data_info_or_list_name``.
       in this case, an attempt is made to download information from the
       DB, where a Dataset called ``<data_info_or_list_name>_info`` should be available
       and have the information normally stored by DataInfo.publish()

    The flag `init_samples` defines whether samples should automatically
    be set up in the costructor.

    If the user needs to modify the `DataDownloader` object before starting
    the training, then `init_samples=False` has to be set.
    In that case, to set up a `DataDownloader`, the user has to call
    `init_samples()`.

    Calling `update_data()`
     - check if new batches are available and download them, if `dynamic` is set to `True`
     - shuffle the dataset if `shuffle` is set to ``True``.

    :param batch_size: Size of batches obtained with __iter__
    :type batch_size: int
    :param dynamic: Whether new batches should be donwnloaded when ``update_data`` is called.
    :type dtnamic: bool
    :param shuffle: whether order of samples has to be shuffled when calling `update_data`
    :type shuffle: bool
    :param data_info_or_list_name: DataInfo object with details about dataset to download, if a string is passed,
                      it is used to download DataInfo data from DB, assuming it was stored with
                      ``list_name=data_info_or_list_name``
    :type data_info_or_list_name: DataInfo | str
    :param list_name: Name of aggregation list used to upload data
    :type list_name: str
    :param cluster: Whether the Orchestrator will be run as a cluster
    :type cluster: bool
    :param address: Address of Redis client as <ip_address>:<port>
    :type address: str
    :param replica_rank: When StaticDataDownloader is used distributedly, indicates
                         the rank of this object
    :type replica_rank: int
    :param num_replicas: When BatchDownlaoder is used distributedly, indicates
                         the total number of ranks
    :type num_replicas: int
    :param verbose: Whether log messages should be printed
    :type verbose: bool
    :param init_samples: whether samples should be initialized in the constructor
    :type init_samples: bool
    :param max_fetch_trials: maximum number of attempts to initialize data
    :type max_fetch_trials: int
    """

    def __init__(
        self,
        data_info_or_list_name,
        batch_size=32,
        dynamic=True,
        shuffle=True,
        cluster=True,
        address=None,
        replica_rank=0,
        num_replicas=1,
        verbose=False,
        init_samples=True,
        max_fetch_trials=-1,
    ):
        self.address = address
        self.cluster = cluster
        self.verbose = verbose
        self.samples = None
        self.targets = None
        self.num_samples = 0
        self.indices = np.arange(0)
        self.shuffle = shuffle
        self.dynamic = dynamic
        self.batch_size = batch_size
        if isinstance(data_info_or_list_name, DataInfo):
            self._info = data_info_or_list_name
        elif isinstance(data_info_or_list_name, str):
            self._info = DataInfo(list_name=data_info_or_list_name)
            client = Client(self.address, self.cluster)
            self._info.download(client)
        else:
            raise TypeError("data_info_or_list_name must be either DataInfo or str")
        self.client = None
        sskeyin = environ.get("SSKEYIN", "")
        self.uploader_keys = sskeyin.split(",")

        self.set_replica_parameters(replica_rank, num_replicas)

        if init_samples:
            self.init_samples(max_fetch_trials)

    def log(self, message):
        if self.verbose:
            logger.info(message)

    def set_replica_parameters(self, replica_rank, num_replicas):
        self.replica_rank = replica_rank
        self.num_replicas = num_replicas
        self.next_indices = [self.replica_rank] * max(1, len(self.uploader_keys))

    @property
    def autoencoding(self):
        return self.sample_name == self.target_name

    @property
    def list_name(self):
        return self._info.list_name

    @property
    def sample_name(self):
        return self._info.sample_name

    @property
    def target_name(self):
        return self._info.target_name

    @property
    def num_classes(self):
        return self._info.num_classes

    @property
    def need_targets(self):
        """Compute if targets have to be downloaded.

        :return: Whether targets (or labels) should be downloaded
        :rtype: bool
        """
        return self.target_name and not self.autoencoding

    def __len__(self):
        length = int(np.floor(self.num_samples / self.batch_size))
        return length

    def __iter__(self):

        self.update_data()
        # Generate data
        if len(self) < 1:
            msg = "Not enough samples in generator for one batch. "
            msg += "Please run init_samples() or initialize generator with init_samples=True"
            raise ValueError(msg)

        _calc_indices = lambda index: self.indices[
            index * self.batch_size : (index + 1) * self.batch_size
        ]
        yield from (
            self._data_generation(_calc_indices(idx)) for idx in range(len(self))
        )

    def init_samples(self, init_trials=-1):
        """Initialize samples (and targets, if needed).

        A new attempt to download samples will be made every ten seconds, for ``init_trials``
        times.
        :param init_trials: maximum number of attempts to fetch data
        :type init_trials: int
        """
        self.client = Client(self.address, self.cluster)

        num_trials = 0
        max_trials = init_trials or -1
        while not len(self) and num_trials != max_trials:
            self._update_samples_and_targets()
            self.log(
                "DataLoader could not download samples, will try again in 10 seconds"
            )
            time.sleep(10)
            num_trials += 1

        if not len(self):
            raise SSInternalError(
                "Could not download samples in given number of trials"
            )
        if self.shuffle:
            np.random.shuffle(self.indices)

    def _data_exists(self, batch_name, target_name):

        if self.need_targets:
            return self.client.tensor_exists(batch_name) and self.client.tensor_exists(
                target_name
            )
        else:
            return self.client.tensor_exists(batch_name)

    def _add_samples(self, indices):

        if self.num_replicas == 1:
            datasets: list[Dataset] = self.client.get_dataset_list_range(
                self.list_name, start_index=indices[0], end_index=indices[-1]
            )
        else:
            datasets: list[Dataset] = []
            for idx in indices:
                datasets += self.client.get_dataset_list_range(
                    self.list_name, start_index=idx, end_index=idx
                )

        if self.samples is None:
            self.samples = datasets[0].get_tensor(self.sample_name)
            if self.need_targets:
                self.targets = datasets[0].get_tensor(self.target_name)

            if len(datasets) > 1:
                datasets = datasets[1:]

        for dataset in datasets:
            self.samples = np.concatenate(
                (self.samples, dataset.get_tensor(self.sample_name))
            )
            if self.need_targets:
                self.targets = np.concatenate(
                    (self.targets, dataset.get_tensor(self.target_name))
                )

        self.num_samples = self.samples.shape[0]
        self.indices = np.arange(self.num_samples)
        self.log(f"New dataset size: {self.num_samples}, batches: {len(self)}")

    def _update_samples_and_targets(self):
        if not self.client:
            self.client = Client(self.address, self.cluster)
            
        self.log(f"Rank {self.replica_rank} out of {self.num_replicas} replicas")

        for uploader_idx, uploader_key in enumerate(self.uploader_keys):
            if uploader_key:
                self.client.use_list_ensemble_prefix(True)
                self.client.set_data_source(uploader_key)

            list_length = self.client.get_list_length(self.list_name)

            # Strictly greater, because next_index is 0-based
            if list_length > self.next_indices[uploader_idx]:
                indices = [
                    idx
                    for idx in range(
                        self.next_indices[uploader_idx], list_length, self.num_replicas
                    )
                ]
                self._add_samples(indices)
                self.next_indices[uploader_idx] = indices[-1] + self.num_replicas

    def update_data(self):
        if self.dynamic:
            self._update_samples_and_targets()
        if self.shuffle:
            np.random.shuffle(self.indices)

    def _data_generation(self, indices):
        # Initialization
        x = self.samples[indices]

        if self.need_targets:
            y = self.targets[indices]
        elif self.autoencoding:
            y = x
        else:
            return x

        return x, y



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re

from smartsim.error.errors import SmartSimError

from ...error import ParameterWriterError
from ...log import get_logger

logger = get_logger(__name__)


class ModelWriter:
    def __init__(self):
        self.tag = ";"
        self.regex = "(;[^;]+;)"
        self.lines = []

    def set_tag(self, tag, regex=None):
        """Set the tag for the modelwriter to search for within
           tagged files attached to an entity.

        :param tag: tag for the modelwriter to search for,
                    defaults to semi-colon e.g. ";"
        :type tag: str
        :param regex: full regex for the modelwriter to search for,
                     defaults to "(;.+;)"
        :type regex: str, optional
        """
        if regex:
            self.regex = regex
        else:
            self.tag = tag
            self.regex = "".join(("(", tag, ".+", tag, ")"))

    def configure_tagged_model_files(
        self, tagged_files, params, make_missing_tags_fatal=False
    ):
        """Read, write and configure tagged files attached to a Model
           instance.

        :param tagged_files: list of paths to tagged files
        :type model: list[str]
        :param params: model parameters
        :type params: dict[str, str]
        :param make_missing_tags_fatal: blow up if a tag is missing
        :type make_missing_tags_fatal: bool
        """
        for tagged_file in tagged_files:
            self._set_lines(tagged_file)
            self._replace_tags(params, make_missing_tags_fatal)
            self._write_changes(tagged_file)

    def _set_lines(self, file_path):
        """Set the lines for the modelwrtter to iterate over

        :param file_path: path to the newly created and tagged file
        :type file_path: str
        :raises ParameterWriterError: if the newly created file cannot be read
        """
        try:
            fp = open(file_path, "r+")
            self.lines = fp.readlines()
            fp.close()
        except (IOError, OSError) as e:
            raise ParameterWriterError(file_path) from e

    def _write_changes(self, file_path):
        """Write the ensemble-specific changes

        :raises ParameterWriterError: if the newly created file cannot be read
        """
        try:
            fp = open(file_path, "w+")
            for line in self.lines:
                fp.write(line)
            fp.close()
        except (IOError, OSError) as e:
            raise ParameterWriterError(file_path, read=False) from e

    def _replace_tags(self, params, make_fatal=False):
        """Replace the tagged within the tagged file attached to this
           model. The tag defaults to ";"

        :param model: The model instance
        :type model: Model
        :param make_fatal: (Optional) Set to True to force a fatal error if a tag is not matched
        :type make_fatal: bool
        """
        edited = []
        unused_tags = {}
        for i, line in enumerate(self.lines):
            search = re.search(self.regex, line)
            if search:
                while search:
                    tagged_line = search.group(0)
                    previous_value = self._get_prev_value(tagged_line)
                    if self._is_ensemble_spec(tagged_line, params):
                        new_val = str(params[previous_value])
                        new_line = re.sub(self.regex, new_val, line, 1)
                        search = re.search(self.regex, new_line)
                        if not search:
                            edited.append(new_line)
                        else:
                            line = new_line

                    # if a tag is found but is not in this model's configurations
                    # put in placeholder value
                    else:
                        tag = tagged_line.split(self.tag)[1]
                        if tag not in unused_tags:
                            unused_tags[tag] = []
                        unused_tags[tag].append(i + 1)
                        edited.append(re.sub(self.regex, previous_value, line))
                        search = False  # Move on to the next tag
            else:
                edited.append(line)
        for tag in unused_tags:
            missing_tag_message = (
                f"Unused tag {tag} on line(s): {str(unused_tags[tag])}"
            )
            if make_fatal:
                raise SmartSimError(missing_tag_message)
            else:
                logger.warning(missing_tag_message)
        self.lines = edited

    def _is_ensemble_spec(self, tagged_line, model_params):
        split_tag = tagged_line.split(self.tag)
        prev_val = split_tag[1]
        if prev_val in model_params.keys():
            return True
        return False

    def _get_prev_value(self, tagged_line):
        split_tag = tagged_line.split(self.tag)
        return split_tag[1]



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .._core.utils.helpers import is_valid_cmd
from ..error import SmartSimError
from ..wlm import detect_launcher
from . import *


def create_batch_settings(
    launcher, nodes=None, time="", queue=None, account=None, batch_args=None, **kwargs
):
    """Create a ``BatchSettings`` instance

    See Experiment.create_batch_settings for details

    :param launcher: launcher for this experiment, if set to 'auto',
                     an attempt will be made to find an available launcher on the system
    :type launcher: str
    :param nodes: number of nodes for batch job, defaults to 1
    :type nodes: int, optional
    :param time: length of batch job, defaults to ""
    :type time: str, optional
    :param queue: queue or partition (if slurm), defaults to ""
    :type queue: str, optional
    :param account: user account name for batch system, defaults to ""
    :type account: str, optional
    :param batch_args: additional batch arguments, defaults to None
    :type batch_args: dict[str, str], optional
    :return: a newly created BatchSettings instance
    :rtype: BatchSettings
    :raises SmartSimError: if batch creation fails
    """
    # all supported batch class implementations
    by_launcher = {
        "cobalt": CobaltBatchSettings,
        "pbs": QsubBatchSettings,
        "slurm": SbatchSettings,
        "lsf": BsubBatchSettings,
    }

    if launcher == "auto":
        launcher = detect_launcher()

    if launcher == "local":
        raise SmartSimError("Local launcher does not support batch workloads")

    # detect the batch class to use based on the launcher provided by
    # the user
    try:
        batch_class = by_launcher[launcher]
        batch_settings = batch_class(
            nodes=nodes,
            time=time,
            batch_args=batch_args,
            queue=queue,
            account=account,
            **kwargs,
        )
        return batch_settings

    except KeyError:
        raise SmartSimError(
            f"User attempted to make batch settings for unsupported launcher {launcher}"
        ) from None


def create_run_settings(
    launcher,
    exe,
    exe_args=None,
    run_command="auto",
    run_args=None,
    env_vars=None,
    container=None,
    **kwargs,
):
    """Create a ``RunSettings`` instance.

    See Experiment.create_run_settings docstring for more details

    :param launcher: launcher to create settings for, if set to 'auto',
                     an attempt will be made to find an available launcher on the system
    :type launcher: str
    :param run_command: command to run the executable
    :type run_command: str
    :param exe: executable to run
    :type exe: str
    :param exe_args: arguments to pass to the executable
    :type exe_args: list[str], optional
    :param run_args: arguments to pass to the ``run_command``
    :type run_args: list[str], optional
    :param env_vars: environment variables to pass to the executable
    :type env_vars: dict[str, str], optional
    :return: the created ``RunSettings``
    :rtype: RunSettings
    :raises SmartSimError: if run_command=="auto" and detection fails
    """
    # all supported RunSettings child classes
    supported = {
        "aprun": AprunSettings,
        "srun": SrunSettings,
        "mpirun": MpirunSettings,
        "mpiexec": MpiexecSettings,
        "orterun": OrterunSettings,
        "jsrun": JsrunSettings,
    }

    # run commands supported by each launcher
    # in order of suspected user preference
    by_launcher = {
        "slurm": ["srun", "mpirun", "mpiexec"],
        "pbs": ["aprun", "mpirun", "mpiexec"],
        "cobalt": ["aprun", "mpirun", "mpiexec"],
        "lsf": ["jsrun", "mpirun", "mpiexec"],
    }

    if launcher == "auto":
        launcher = detect_launcher()

    def _detect_command(launcher):
        if launcher in by_launcher:
            for cmd in by_launcher[launcher]:
                if is_valid_cmd(cmd):
                    return cmd
        msg = f"Could not automatically detect a run command to use for launcher {launcher}"
        msg += f"\nSearched for and could not find the following commands: {by_launcher[launcher]}"
        raise SmartSimError(msg)

    if run_command:
        run_command = run_command.lower()
    launcher = launcher.lower()

    # detect run_command automatically for all but local launcher
    if run_command == "auto":
        # no auto detection for local, revert to false
        if launcher == "local":
            run_command = None
        else:
            run_command = _detect_command(launcher)

    # if user specified and supported or auto detection worked
    if run_command and run_command in supported:
        return supported[run_command](
            exe, exe_args, run_args, env_vars, container=container, **kwargs
        )

    # 1) user specified and not implementation in SmartSim
    # 2) user supplied run_command=None
    # 3) local launcher being used and default of "auto" was passed.
    return RunSettings(
        exe, exe_args, run_command, run_args, env_vars, container=container
    )



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from tensorflow import keras

from smartsim.ml import DataDownloader


class _TFDataGenerationCommon(DataDownloader, keras.utils.Sequence):
    def __getitem__(self, index):
        if len(self) < 1:
            msg = "Not enough samples in generator for one batch. "
            msg += "Please run init_samples() or initialize generator with init_samples=True"
            raise ValueError(msg)
        # Generate indices of the batch
        indices = self.indices[index * self.batch_size : (index + 1) * self.batch_size]

        # Generate data
        x, y = self._data_generation(indices)

        if y is not None:
            return x, y
        else:
            return x

    def on_epoch_end(self):
        """Callback called at the end of each training epoch

        If `self.shuffle` is set to `True`, data is shuffled.
        """
        if self.shuffle:
            np.random.shuffle(self.indices)

    def _data_generation(self, indices):
        # Initialization
        x = self.samples[indices]

        if self.need_targets:
            y = self.targets[indices]
            if self.num_classes is not None:
                y = keras.utils.to_categorical(y, num_classes=self.num_classes)
        elif self.autoencoding:
            y = x
        else:
            return x

        return x, y


class StaticDataGenerator(_TFDataGenerationCommon):
    """A class to download a dataset from the DB.

    Details about parameters and features of this class can be found
    in the documentation of ``DataDownloader``, of which it is just
    a TensorFlow-specialized sub-class with dynamic=False.
    """

    def __init__(self, **kwargs):
        dynamic = kwargs.pop("dynamic", False)
        kwargs["dynamic"] = False
        super().__init__(**kwargs)
        if dynamic:
            self.log(
                "Static data generator cannot be started with dynamic=True, setting it to False"
            )


class DynamicDataGenerator(_TFDataGenerationCommon):
    """A class to download batches from the DB.

    Details about parameters and features of this class can be found
    in the documentation of ``DataDownloader``, of which it is just
    a TensorFlow-specialized sub-class with dynamic=True.
    """

    def __init__(self, **kwargs):
        dynamic = kwargs.pop("dynamic", True)
        kwargs["dynamic"] = True
        super().__init__(**kwargs)
        if not dynamic:
            self.log(
                "Dynamic data generator cannot be started with dynamic=False, setting it to True"
            )

    def on_epoch_end(self):
        """Callback called at the end of each training epoch

        Update data (the DB is queried for new batches) and
        if `self.shuffle` is set to `True`, data is also shuffled.
        """
        self.update_data()



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import warnings

from .._core.utils.helpers import cat_arg_and_value, init_default
from ..error import EntityExistsError, SSUnsupportedError
from .dbobject import DBModel, DBScript
from .entity import SmartSimEntity
from .files import EntityFiles


class Model(SmartSimEntity):
    def __init__(
        self, name, params, path, run_settings, params_as_args=None, batch_settings=None
    ):
        """Initialize a ``Model``

        :param name: name of the model
        :type name: str
        :param params: model parameters for writing into configuration files or
                       to be passed as command line arguments to executable.
        :type params: dict
        :param path: path to output, error, and configuration files
        :type path: str
        :param run_settings: launcher settings specified in the experiment
        :type run_settings: RunSettings
        :param params_as_args: list of parameters which have to be
                               interpreted as command line arguments to
                               be added to run_settings
        :type params_as_args: list[str]
        :param batch_settings: Launcher settings for running the individual
                               model as a batch job, defaults to None
        :type batch_settings: BatchSettings | None
        """
        super().__init__(name, path, run_settings)
        self.params = params
        self.params_as_args = params_as_args
        self.incoming_entities = []
        self._key_prefixing_enabled = False
        self.batch_settings = batch_settings
        self._db_models = []
        self._db_scripts = []
        self.files = None

    @property
    def colocated(self):
        """Return True if this Model will run with a colocated Orchestrator"""
        return bool(self.run_settings.colocated_db_settings)

    def register_incoming_entity(self, incoming_entity):
        """Register future communication between entities.

        Registers the named data sources that this entity
        has access to by storing the key_prefix associated
        with that entity

        :param incoming_entity: The entity that data will be received from
        :type incoming_entity: SmartSimEntity
        :raises SmartSimError: if incoming entity has already been registered
        """
        if incoming_entity.name in [
            in_entity.name for in_entity in self.incoming_entities
        ]:
            raise EntityExistsError(
                f"'{incoming_entity.name}' has already "
                + "been registered as an incoming entity"
            )

        self.incoming_entities.append(incoming_entity)

    def enable_key_prefixing(self):
        """If called, the entity will prefix its keys with its own model name"""
        self._key_prefixing_enabled = True

    def disable_key_prefixing(self):
        """If called, the entity will not prefix its keys with its own model name"""
        self._key_prefixing_enabled = False

    def query_key_prefixing(self):
        """Inquire as to whether this entity will prefix its keys with its name"""
        return self._key_prefixing_enabled

    def attach_generator_files(self, to_copy=None, to_symlink=None, to_configure=None):
        """Attach files to an entity for generation

        Attach files needed for the entity that, upon generation,
        will be located in the path of the entity.

        During generation, files "to_copy" are copied into
        the path of the entity, and files "to_symlink" are
        symlinked into the path of the entity.

        Files "to_configure" are text based model input files where
        parameters for the model are set. Note that only models
        support the "to_configure" field. These files must have
        fields tagged that correspond to the values the user
        would like to change. The tag is settable but defaults
        to a semicolon e.g. THERMO = ;10;

        :param to_copy: files to copy, defaults to []
        :type to_copy: list, optional
        :param to_symlink: files to symlink, defaults to []
        :type to_symlink: list, optional
        :param to_configure: input files with tagged parameters, defaults to []
        :type to_configure: list, optional
        """
        to_copy = init_default([], to_copy, (list, str))
        to_symlink = init_default([], to_symlink, (list, str))
        to_configure = init_default([], to_configure, (list, str))
        self.files = EntityFiles(to_configure, to_copy, to_symlink)

    def colocate_db(self, *args, **kwargs):
        """An alias for ``Model.colocate_db_tcp``"""
        warnings.warn(
            (
                "`colocate_db` has been deprecated and will be removed in a \n"
                "future release. Please use `colocate_db_tcp` or `colocate_db_uds`."
            ),
            category=DeprecationWarning,
        )
        self.colocate_db_tcp(*args, **kwargs)

    def colocate_db_uds(
        self,
        unix_socket="/tmp/redis.socket",
        socket_permissions=755,
        db_cpus=1,
        limit_app_cpus=True,
        debug=False,
        **kwargs,
    ):
        """Colocate an Orchestrator instance with this Model over UDS.

        This method will initialize settings which add an unsharded
        database to this Model instance. Only this Model will be able to communicate
        with this colocated database by using Unix Domain sockets.

        Extra parameters for the db can be passed through kwargs. This includes
        many performance, caching and inference settings.

        .. highlight:: python
        .. code-block:: python

            example_kwargs = {
                "maxclients": 100000,
                "threads_per_queue": 1,
                "inter_op_threads": 1,
                "intra_op_threads": 1,
                "server_threads": 2 # keydb only
            }

        Generally these don't need to be changed.

        :param unix_socket: path to where the socket file will be created
        :type unix_socket: str, optional
        :param socket_permissions: permissions for the socketfile
        :type socket_permissions: int, optional
        :param db_cpus: number of cpus to use for orchestrator, defaults to 1
        :type db_cpus: int, optional
        :param limit_app_cpus: whether to limit the number of cpus used by the app, defaults to True
        :type limit_app_cpus: bool, optional
        :param debug: launch Model with extra debug information about the co-located db
        :type debug: bool, optional
        :param kwargs: additional keyword arguments to pass to the orchestrator database
        :type kwargs: dict, optional
        """

        uds_options = {
            "unix_socket": unix_socket,
            "socket_permissions": socket_permissions,
            "port": 0,  # This is hardcoded to 0 as recommended by redis for UDS
        }
        common_options = {
            "cpus": db_cpus,
            "limit_app_cpus": limit_app_cpus,
            "debug": debug,
        }
        self._set_colocated_db_settings(uds_options, common_options, **kwargs)

    def colocate_db_tcp(
        self,
        port=6379,
        ifname="lo",
        db_cpus=1,
        limit_app_cpus=True,
        debug=False,
        **kwargs,
    ):
        """Colocate an Orchestrator instance with this Model over TCP/IP.

        This method will initialize settings which add an unsharded
        database to this Model instance. Only this Model will be able to communicate
        with this colocated database by using the loopback TCP interface.

        Extra parameters for the db can be passed through kwargs. This includes
        many performance, caching and inference settings.

        .. highlight:: python
        .. code-block:: python

            ex. kwargs = {
                maxclients: 100000,
                threads_per_queue: 1,
                inter_op_threads: 1,
                intra_op_threads: 1,
                server_threads: 2 # keydb only
            }

        Generally these don't need to be changed.

        :param port: port to use for orchestrator database, defaults to 6379
        :type port: int, optional
        :param ifname: interface to use for orchestrator, defaults to "lo"
        :type ifname: str, optional
        :param db_cpus: number of cpus to use for orchestrator, defaults to 1
        :type db_cpus: int, optional
        :param limit_app_cpus: whether to limit the number of cpus used by the app, defaults to True
        :type limit_app_cpus: bool, optional
        :param debug: launch Model with extra debug information about the co-located db
        :type debug: bool, optional
        :param kwargs: additional keyword arguments to pass to the orchestrator database
        :type kwargs: dict, optional

        """

        tcp_options = {"port": port, "ifname": ifname}
        common_options = {
            "cpus": db_cpus,
            "limit_app_cpus": limit_app_cpus,
            "debug": debug,
        }
        self._set_colocated_db_settings(tcp_options, common_options, **kwargs)

    def _set_colocated_db_settings(self, connection_options, common_options, **kwargs):
        """
        Ingest the connection-specific options (UDS/TCP) and set the final settings
        for the co-located database
        """

        if hasattr(self.run_settings, "mpmd") and len(self.run_settings.mpmd) > 0:
            raise SSUnsupportedError(
                "Models co-located with databases cannot be run as a mpmd workload"
            )

        if hasattr(self.run_settings, "_prep_colocated_db"):
            self.run_settings._prep_colocated_db(common_options["cpus"])

        # TODO list which db settings can be extras
        colo_db_config = {}
        colo_db_config.update(connection_options)
        colo_db_config.update(common_options)
        # redisai arguments for inference settings
        colo_db_config["rai_args"] = {
            "threads_per_queue": kwargs.get("threads_per_queue", None),
            "inter_op_parallelism": kwargs.get("inter_op_parallelism", None),
            "intra_op_parallelism": kwargs.get("intra_op_parallelism", None),
        }
        colo_db_config["extra_db_args"] = dict(
            [
                (k, str(v))
                for k, v in kwargs.items()
                if k not in colo_db_config["rai_args"]
            ]
        )

        self._check_db_objects_colo()
        colo_db_config["db_models"] = self._db_models
        colo_db_config["db_scripts"] = self._db_scripts

        self.run_settings.colocated_db_settings = colo_db_config

    def params_to_args(self):
        """Convert parameters to command line arguments and update run settings."""
        for param in self.params_as_args:
            if not param in self.params:
                raise ValueError(
                    f"Tried to convert {param} to command line argument "
                    + f"for Model {self.name}, but its value was not found in model params"
                )
            if self.run_settings is None:
                raise ValueError(
                    f"Tried to configure command line parameter for Model {self.name}, "
                    + "but no RunSettings are set."
                )
            self.run_settings.add_exe_args(cat_arg_and_value(param, self.params[param]))

    def add_ml_model(
        self,
        name,
        backend,
        model=None,
        model_path=None,
        device="CPU",
        devices_per_node=1,
        batch_size=0,
        min_batch_size=0,
        tag="",
        inputs=None,
        outputs=None,
    ):
        """A TF, TF-lite, PT, or ONNX model to load into the DB at runtime

        Each ML Model added will be loaded into an
        orchestrator (converged or not) prior to the execution
        of this Model instance

        One of either model (in memory representation) or model_path (file)
        must be provided

        :param name: key to store model under
        :type name: str
        :param model: model in memory
        :type model: byte string, optional
        :param model_path: serialized model
        :type model_path: file path to model
        :param backend: name of the backend (TORCH, TF, TFLITE, ONNX)
        :type backend: str
        :param device: name of device for execution, defaults to "CPU"
        :type device: str, optional
        :param batch_size: batch size for execution, defaults to 0
        :type batch_size: int, optional
        :param min_batch_size: minimum batch size for model execution, defaults to 0
        :type min_batch_size: int, optional
        :param tag: additional tag for model information, defaults to ""
        :type tag: str, optional
        :param inputs: model inputs (TF only), defaults to None
        :type inputs: list[str], optional
        :param outputs: model outupts (TF only), defaults to None
        :type outputs: list[str], optional
        """
        db_model = DBModel(
            name=name,
            backend=backend,
            model=model,
            model_file=model_path,
            device=device,
            devices_per_node=devices_per_node,
            batch_size=batch_size,
            min_batch_size=min_batch_size,
            tag=tag,
            inputs=inputs,
            outputs=outputs,
        )
        self._append_db_model(db_model)

    def add_script(
        self, name, script=None, script_path=None, device="CPU", devices_per_node=1
    ):
        """TorchScript to launch with this Model instance

        Each script added to the model will be loaded into an
        orchestrator (converged or not) prior to the execution
        of this Model instance

        Device selection is either "GPU" or "CPU". If many devices are
        present, a number can be passed for specification e.g. "GPU:1".

        Setting ``devices_per_node=N``, with N greater than one will result
        in the model being stored in the first N devices of type ``device``.

        One of either script (in memory string representation) or script_path (file)
        must be provided

        :param name: key to store script under
        :type name: str
        :param script: TorchScript code
        :type script: str, optional
        :param script_path: path to TorchScript code
        :type script_path: str, optional
        :param device: device for script execution, defaults to "CPU"
        :type device: str, optional
        :param devices_per_node: number of devices on each host
        :type devices_per_node: int
        """
        db_script = DBScript(
            name=name,
            script=script,
            script_path=script_path,
            device=device,
            devices_per_node=devices_per_node,
        )
        self._append_db_script(db_script)

    def add_function(self, name, function=None, device="CPU", devices_per_node=1):
        """TorchScript function to launch with this Model instance

        Each script function to the model will be loaded into a
        non-converged orchestrator prior to the execution
        of this Model instance.

        For converged orchestrators, the :meth:`add_script` method should be used.

        Device selection is either "GPU" or "CPU". If many devices are
        present, a number can be passed for specification e.g. "GPU:1".

        Setting ``devices_per_node=N``, with N greater than one will result
        in the model being stored in the first N devices of type ``device``.

        :param name: key to store function under
        :type name: str
        :param script: TorchScript code
        :type script: str or byte string, optional
        :param script_path: path to TorchScript code
        :type script_path: str, optional
        :param device: device for script execution, defaults to "CPU"
        :type device: str, optional
        :param devices_per_node: number of devices on each host
        :type devices_per_node: int
        """
        db_script = DBScript(
            name=name, script=function, device=device, devices_per_node=devices_per_node
        )
        self._append_db_script(db_script)

    def __eq__(self, other):
        if self.name == other.name:
            return True
        return False

    def __str__(self):  # pragma: no cover
        entity_str = "Name: " + self.name + "\n"
        entity_str += "Type: " + self.type + "\n"
        entity_str += str(self.run_settings) + "\n"
        if self._db_models:
            entity_str += "DB Models: \n" + str(len(self._db_models)) + "\n"
        if self._db_scripts:
            entity_str += "DB Scripts: \n" + str(len(self._db_scripts)) + "\n"
        return entity_str

    def _append_db_model(self, db_model):
        if not db_model.is_file and self.colocated:
            err_msg = "ML model can not be set from memory for colocated databases.\n"
            err_msg += (
                f"Please store the ML model named {db_model.name} in binary format "
            )
            err_msg += "and add it to the SmartSim Model as file."
            raise SSUnsupportedError(err_msg)

        self._db_models.append(db_model)

    def _append_db_script(self, db_script):
        if db_script.func and self.colocated:
            if not isinstance(db_script.func, str):
                err_msg = (
                    "Functions can not be set from memory for colocated databases.\n"
                )
                err_msg += f"Please convert the function named {db_script.name} to a string or store "
                err_msg += "it as a text file and add it to the SmartSim Model with add_script."
                raise SSUnsupportedError(err_msg)
        self._db_scripts.append(db_script)

    def _check_db_objects_colo(self):

        for db_model in self._db_models:
            if not db_model.is_file:
                err_msg = (
                    "ML model can not be set from memory for colocated databases.\n"
                )
                err_msg += (
                    f"Please store the ML model named {db_model.name} in binary format "
                )
                err_msg += "and add it to the SmartSim Model as file."
                raise SSUnsupportedError(err_msg)

        for db_script in self._db_scripts:
            if db_script.func:
                if not isinstance(db_script.func, str):
                    err_msg = "Functions can not be set from memory for colocated databases.\n"
                    err_msg += f"Please convert the function named {db_script.name} to a string or store it "
                    err_msg += "as a text file and add it to the SmartSim Model with add_script."
                    raise SSUnsupportedError(err_msg)



# BSD 2-Clause License #
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .._core.utils.helpers import expand_exe_path, fmt_dict, init_default, is_valid_cmd
from ..log import get_logger

logger = get_logger(__name__)


class RunSettings:
    def __init__(
        self,
        exe,
        exe_args=None,
        run_command="",
        run_args=None,
        env_vars=None,
        container=None,
        **kwargs,
    ):
        """Run parameters for a ``Model``

        The base ``RunSettings`` class should only be used with the `local`
        launcher on single node, workstations, or laptops.

        If no ``run_command`` is specified, the executable will be launched
        locally.

        ``run_args`` passed as a dict will be interpreted literally for
        local ``RunSettings`` and added directly to the ``run_command``
        e.g. run_args = {"-np": 2} will be "-np 2"

        Example initialization

        .. highlight:: python
        .. code-block:: python

            rs = RunSettings("echo", "hello", "mpirun", run_args={"-np": "2"})

        :param exe: executable to run
        :type exe: str
        :param exe_args: executable arguments, defaults to None
        :type exe_args: str | list[str], optional
        :param run_command: launch binary (e.g. "srun"), defaults to empty str
        :type run_command: str, optional
        :param run_args: arguments for run command (e.g. `-np` for `mpiexec`), defaults to None
        :type run_args: dict[str, str], optional
        :param env_vars: environment vars to launch job with, defaults to None
        :type env_vars: dict[str, str], optional
        :param container: container type for workload (e.g. "singularity"), defaults to None
        :type container: Container, optional
        """
        # Do not expand executable if running within a container
        if container:
            self.exe = [exe]
        else:
            self.exe = [expand_exe_path(exe)]

        self.exe_args = self._set_exe_args(exe_args)
        self.run_args = init_default({}, run_args, dict)
        self.env_vars = init_default({}, env_vars, dict)
        self.container = container
        self._run_command = run_command
        self.in_batch = False
        self.colocated_db_settings = None

    # To be overwritten by subclasses. Set of reserved args a user cannot change
    reserved_run_args = set()  # type: set[str]

    def set_nodes(self, nodes):
        """Set the number of nodes

        :param nodes: number of nodes to run with
        :type nodes: int
        """
        logger.warning(
            (
                "Node specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_tasks(self, tasks):
        """Set the number of tasks to launch

        :param tasks: number of tasks to launch
        :type tasks: int
        """
        logger.warning(
            (
                "Task specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_tasks_per_node(self, tasks_per_node):
        """Set the number of tasks per node

        :param tasks_per_node: number of tasks to launch per node
        :type tasks_per_node: int
        """
        logger.warning(
            (
                "Task per node specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_task_map(self, task_mapping):
        """Set a task mapping

        :param task_mapping: task mapping
        :type task_mapping: str
        """
        logger.warning(
            (
                "Task mapping specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_cpus_per_task(self, cpus_per_task):
        """Set the number of cpus per task

        :param cpus_per_task: number of cpus per task
        :type cpus_per_task: int
        """
        logger.warning(
            (
                "CPU per node specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_hostlist(self, host_list):
        """Specify the hostlist for this job

        :param host_list: hosts to launch on
        :type host_list: str | list[str]
        """
        logger.warning(
            (
                "Hostlist specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_hostlist_from_file(self, file_path):
        """Use the contents of a file to specify the hostlist for this job

        :param file_path: Path to the hostlist file
        :type file_path: str
        """
        logger.warning(
            (
                "Hostlist from file specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_excluded_hosts(self, host_list):
        """Specify a list of hosts to exclude for launching this job

        :param host_list: hosts to exclude
        :type host_list: str | list[str]
        """
        logger.warning(
            (
                "Excluded host list specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_cpu_bindings(self, bindings):
        """Set the cores to which MPI processes are bound

        :param bindings: List specifing the cores to which MPI processes are bound
        :type bindings: list[int] | int
        """
        logger.warning(
            (
                "CPU binding specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_memory_per_node(self, memory_per_node):
        """Set the amount of memory required per node in megabytes

        :param memory_per_node: Number of megabytes per node
        :type memory_per_node: int
        """
        logger.warning(
            (
                "Memory per node specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_verbose_launch(self, verbose):
        """Set the job to run in verbose mode

        :param verbose: Whether the job should be run verbosely
        :type verbose: bool
        """
        logger.warning(
            (
                "Verbose specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_quiet_launch(self, quiet):
        """Set the job to run in quiet mode

        :param quiet: Whether the job should be run quietly
        :type quiet: bool
        """
        logger.warning(
            (
                "Quiet specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_broadcast(self, dest_path=None):
        """Copy executable file to allocated compute nodes

        :param dest_path: Path to copy an executable file
        :type dest_path: str | None
        """
        logger.warning(
            (
                "Broadcast specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_time(self, hours=0, minutes=0, seconds=0):
        """Automatically format and set wall time

        :param hours: number of hours to run job
        :type hours: int
        :param minutes: number of minutes to run job
        :type minutes: int
        :param seconds: number of seconds to run job
        :type seconds: int
        """
        return self.set_walltime(
            self._fmt_walltime(int(hours), int(minutes), int(seconds))
        )

    def _fmt_walltime(self, hours, minutes, seconds):
        """Convert hours, minutes, and seconds into valid walltime format

        By defualt the formatted wall time is the total number of seconds.

        :param hours: number of hours to run job
        :type hours: int
        :param minutes: number of minutes to run job
        :type minutes: int
        :param seconds: number of seconds to run job
        :type seconds: int
        :returns: Formatted walltime
        :rtype: str
        """
        time_ = hours * 3600
        time_ += minutes * 60
        time_ += seconds
        return str(time_)

    def set_walltime(self, walltime):
        """Set the formatted walltime

        :param walltime: Time in format required by launcher``
        :type walltime: str
        """
        logger.warning(
            (
                "Walltime specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_binding(self, binding):
        """Set binding

        :param binding: Binding
        :type binding: str
        """
        logger.warning(
            (
                "binding specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def set_mpmd_preamble(self, preamble_lines):
        """Set preamble to a file to make a job MPMD

        :param preamble_lines: lines to put at the beginning of a file.
        :type preamble_lines: list[str]
        """
        logger.warning(
            (
                "MPMD preamble specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    def make_mpmd(self, settings):
        """Make job an MPMD job

        :param settings: ``RunSettings`` instance
        :type aprun_settings: RunSettings
        """
        logger.warning(
            (
                "Make MPMD specification not implemented for this "
                f"RunSettings type: {type(self)}"
            )
        )

    @property
    def run_command(self):
        """Return the launch binary used to launch the executable

        Attempt to expand the path to the executable if possible

        :returns: launch binary e.g. mpiexec
        :type: str | None
        """
        cmd = self._run_command

        if cmd:
            if is_valid_cmd(cmd):
                # command is valid and will be expanded
                return expand_exe_path(cmd)
            # command is not valid, so return it as is
            # it may be on the compute nodes but not local machine
            return cmd
        # run without run command
        return None

    def update_env(self, env_vars):
        """Update the job environment variables

        To fully inherit the current user environment, add the
        workload-manager-specific flag to the launch command through the
        :meth:`add_exe_args` method. For example, ``--export=ALL`` for
        slurm, or ``-V`` for PBS/aprun.


        :param env_vars: environment variables to update or add
        :type env_vars: dict[str, Union[str, int, float, bool]]
        :raises TypeError: if env_vars values cannot be coerced to strings
        """
        val_types = (str, int, float, bool)
        # Coerce env_vars values to str as a convenience to user
        for env, val in env_vars.items():
            if not isinstance(val, val_types):
                raise TypeError(
                    f"env_vars[{env}] was of type {type(val)}, not {val_types}"
                )
            else:
                self.env_vars[env] = val

    def add_exe_args(self, args):
        """Add executable arguments to executable

        :param args: executable arguments
        :type args: str | list[str]
        :raises TypeError: if exe args are not strings
        """
        if isinstance(args, str):
            args = args.split()
        for arg in args:
            if not isinstance(arg, str):
                raise TypeError("Executable arguments should be a list of str")
            self.exe_args.append(arg)

    def set(self, arg, value=None, condition=True):
        """Allows users to set individual run arguments.

        A method that allows users to set run arguments after object
        instantiation. Does basic formatting such as stripping leading dashes.
        If the argument has been set previously, this method will log warning
        but ultimately comply.

        Conditional expressions may be passed to the conditional parameter. If the
        expression evaluates to True, the argument will be set. In not an info
        message is logged and no further operation is performed.

        Basic Usage

        .. highlight:: python
        .. code-block:: python

            rs = RunSettings("python")
            rs.set("an-arg", "a-val")
            rs.set("a-flag")
            rs.format_run_args()  # returns ["an-arg", "a-val", "a-flag", "None"]

        Slurm Example with Conditional Setting

        .. highlight:: python
        .. code-block:: python

            import socket

            rs = SrunSettings("echo", "hello")
            rs.set_tasks(1)
            rs.set("exclusive")

            # Only set this argument if condition param evals True
            # Otherwise log and NOP
            rs.set("partition", "debug",
                   condition=socket.gethostname()=="testing-system")

            rs.format_run_args()
            # returns ["exclusive", "None", "partition", "debug"] iff socket.gethostname()=="testing-system"
            # otherwise returns ["exclusive", "None"]

        :param arg: name of the argument
        :type arg: str
        :param value: value of the argument
        :type value: str | None
        :param conditon: set the argument if condition evaluates to True
        :type condition: bool
        """
        if not isinstance(arg, str):
            raise TypeError("Argument name should be of type str")
        if value is not None and not isinstance(value, str):
            raise TypeError("Argument value should be of type str or None")
        arg = arg.strip().lstrip("-")

        if not condition:
            logger.info(f"Could not set argument '{arg}': condition not met")
            return
        if arg in self.reserved_run_args:
            logger.warning(
                (
                    f"Could not set argument '{arg}': "
                    f"it is a reserved arguement of '{type(self).__name__}'"
                )
            )
            return

        if arg in self.run_args and value != self.run_args[arg]:
            logger.warning(f"Overwritting argument '{arg}' with value '{value}'")
        self.run_args[arg] = value

    def _set_exe_args(self, exe_args):
        if exe_args:
            if isinstance(exe_args, str):
                return exe_args.split()
            if isinstance(exe_args, list):
                plain_type = all([isinstance(arg, (str)) for arg in exe_args])
                if not plain_type:
                    nested_type = all(
                        [
                            all([isinstance(arg, (str)) for arg in exe_args_list])
                            for exe_args_list in exe_args
                        ]
                    )
                    if not nested_type:
                        raise TypeError(
                            "Executable arguments were not list of str or str"
                        )
                    else:
                        return exe_args
                return exe_args
            raise TypeError("Executable arguments were not list of str or str")
        else:
            return []

    def format_run_args(self):
        """Return formatted run arguments

        For ``RunSettings``, the run arguments are passed
        literally with no formatting.

        :return: list run arguments for these settings
        :rtype: list[str]
        """
        formatted = []
        for arg, value in self.run_args.items():
            formatted.append(arg)
            formatted.append(str(value))
        return formatted

    def format_env_vars(self):
        """Build environment variable string

        :returns: formatted list of strings to export variables
        :rtype: list[str]
        """
        formatted = []
        for key, val in self.env_vars.items():
            if val is None:
                formatted.append(f"{key}=")
            else:
                formatted.append(f"{key}={val}")
        return formatted

    def __str__(self):  # pragma: no-cover
        string = f"Executable: {self.exe[0]}\n"
        string += f"Executable Arguments: {' '.join((self.exe_args))}"
        if self.run_command:
            string += f"\nRun Command: {self._run_command}"
        if self.run_args:
            string += f"\nRun Arguments:\n{fmt_dict(self.run_args)}"
        if self.colocated_db_settings:
            string += "\nCo-located Database: True"
        return string


class BatchSettings:
    def __init__(self, batch_cmd, batch_args=None, **kwargs):
        self._batch_cmd = batch_cmd
        self.batch_args = init_default({}, batch_args, dict)
        self._preamble = []
        self.set_nodes(kwargs.get("nodes", None))
        self.set_walltime(kwargs.get("time", None))
        self.set_queue(kwargs.get("queue", None))
        self.set_account(kwargs.get("account", None))

    @property
    def batch_cmd(self):
        """Return the batch command

        Tests to see if we can expand the batch command
        path. If we can, then returns the expanded batch
        command. If we cannot, returns the batch command as is.

        :returns: batch command
        :type: str
        """
        if is_valid_cmd(self._batch_cmd):
            return expand_exe_path(self._batch_cmd)
        else:
            return self._batch_cmd

    def set_nodes(self, num_nodes):
        raise NotImplementedError

    def set_hostlist(self, host_list):
        raise NotImplementedError

    def set_queue(self, queue):
        raise NotImplementedError

    def set_walltime(self, walltime):
        raise NotImplementedError

    def set_account(self, account):
        raise NotImplementedError

    def format_batch_args(self):
        raise NotImplementedError

    def set_batch_command(self, command):
        """Set the command used to launch the batch e.g. ``sbatch``

        :param command: batch command
        :type command: str
        """
        self._batch_cmd = command

    def add_preamble(self, lines):
        """Add lines to the batch file preamble. The lines are just
        written (unmodified) at the beginning of the batch file
        (after the WLM directives) and can be used to e.g.
        start virtual environments before running the executables.

        :param line: lines to add to preamble.
        :type line: str or list[str]
        """
        if isinstance(lines, str):
            self._preamble += [lines]
        elif isinstance(lines, list):
            self._preamble += lines
        else:
            raise TypeError("Expected str or List[str] for lines argument")

    def __str__(self):  # pragma: no-cover
        string = f"Batch Command: {self._batch_cmd}"
        if self.batch_args:
            string += f"\nBatch arguments:\n{fmt_dict(self.batch_args)}"
        return string



# BSD 2-Clause License
#
# Copyright (c) 2021-2023 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import socket

import psutil

"""
A handful of useful functions for dealing with networks
"""


def get_ip_from_host(host):
    """Return the IP address for the interconnect.

    :param host: hostname of the compute node e.g. nid00004
    :type host: str
    :returns: ip of host
    :rtype: str
    """
    ip_address = socket.gethostbyname(host)
    return ip_address


# impossible to cover as it's only used in entrypoints
def get_ip_from_interface(interface):  # pragma: no cover
    """Get IPV4 address of a network interface

    :param interface: interface name
    :type interface: str
    :raises ValueError: if the interface does not exist
    :raises ValueError: if interface does not have an IPV4 address
    :return: ip address of interface
    :rtype: str
    """
    net_if_addrs = psutil.net_if_addrs()
    if interface not in net_if_addrs:

        available = list(net_if_addrs.keys())
        raise ValueError(
            f"{interface} is not a valid network interface. "
            f"Valid network interfaces are: {available}"
        )

    for info in net_if_addrs[interface]:
        if info.family == socket.AF_INET:
            return info.address
    raise ValueError(f"interface {interface} doesn't have an IPv4 address")


# impossible to cover as it's only used in entrypoints
def get_lb_interface_name():  # pragma: no cover
    """Use psutil to get loopback interface name"""
    net_if_addrs = list(psutil.net_if_addrs())
    for interface in net_if_addrs:
        if interface.startswith("lo"):
            return interface
    raise OSError("Could not find loopback interface name")


def current_ip(interface="lo"):  # pragma: no cover
    if interface == "lo":
        loopback = get_lb_interface_name()
        return get_ip_from_interface(loopback)
    else:
        return get_ip_from_interface(interface)



# BSD 2-Clause License
#
# Copyright (c) 2021-2023 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import signal
from subprocess import PIPE, STDOUT
from typing import List

import psutil

from smartsim._core.utils.network import current_ip
from smartsim.error import SSInternalError
from smartsim.log import get_logger

logger = get_logger(__name__)

"""
Redis/KeyDB entrypoint script
"""

DBPID = None

# kill is not catchable
SIGNALS = [signal.SIGINT, signal.SIGQUIT, signal.SIGTERM, signal.SIGABRT]


def handle_signal(signo, frame):
    cleanup()


def main(network_interface: str, command: List[str]):
    global DBPID

    try:

        ip_address = current_ip(network_interface)
        cmd = command + [f"--bind {ip_address}"]

        print("-" * 10, "  Running  Command  ", "-" * 10, "\n", flush=True)
        print(f"COMMAND: {' '.join(cmd)}\n", flush=True)
        print(f"IPADDRESS: {ip_address}\n", flush=True)
        print(f"NETWORK: {network_interface}\n", flush=True)
        print("-" * 30, "\n\n", flush=True)

        print("-" * 10, "  Output  ", "-" * 10, "\n\n", flush=True)

        p = psutil.Popen(cmd, stdout=PIPE, stderr=STDOUT)
        DBPID = p.pid

        for line in iter(p.stdout.readline, b""):
            print(line.decode("utf-8").rstrip(), flush=True)
    except Exception as e:
        cleanup()
        raise SSInternalError("Database process starter raised an exception") from e


def cleanup():
    global DBPID
    try:
        logger.debug("Cleaning up database instance")
        # attempt to stop the database process
        db_proc = psutil.Process(DBPID)
        db_proc.terminate()

    except psutil.NoSuchProcess:
        logger.warning("Couldn't find database process to kill.")

    except OSError as e:
        logger.warning(f"Failed to clean up database gracefully: {str(e)}")


if __name__ == "__main__":

    os.environ["PYTHONUNBUFFERED"] = "1"

    parser = argparse.ArgumentParser(
        prefix_chars="+", description="SmartSim Process Launcher"
    )
    parser.add_argument(
        "+ifname", type=str, help="Network Interface name", default="lo"
    )
    parser.add_argument("+command", nargs="+", help="Command to run")
    args = parser.parse_args()

    # make sure to register the cleanup before the start
    # the proecss so our signaller will be able to stop
    # the database process.
    for sig in SIGNALS:
        signal.signal(sig, handle_signal)

    main(args.ifname, args.command)



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
import os
import sys

import coloredlogs

# constants for logging
coloredlogs.DEFAULT_DATE_FORMAT = "%H:%M:%S"
coloredlogs.DEFAULT_LOG_FORMAT = (
    "%(asctime)s %(hostname)s %(name)s[%(process)d] %(levelname)s %(message)s"
)
# optional thread name logging for debugging
# coloredlogs.DEFAULT_LOG_FORMAT = '%(asctime)s [%(threadName)s] %(hostname)s %(name)s[%(process)d] %(levelname)s %(message)s'


def _get_log_level():
    """Get the logging level based on environment variable
       SMARTSIM_LOG_LEVEL.  If not set, default to info.

       Logging levels
         - quiet: Just shows errors and warnings
         - info: Show basic information and errors (default)
         - debug: Shows info, errors and user debug information
         - developer: Shows everything happening during execution
                      extremely verbose logging.

    :return: Log level for coloredlogs
    :rtype: str
    """
    log_level = os.environ.get("SMARTSIM_LOG_LEVEL", "info").lower()
    if log_level == "quiet":
        return "warning"
    if log_level == "info":
        return "info"
    if log_level == "debug":
        return "debug"
    # extremely verbose logging used internally
    if log_level == "developer":
        return "debug"
    return "info"


def get_logger(name, log_level=None, fmt=None):
    """Return a logger instance

    levels:
        - quiet
        - info
        - debug
        - developer

    examples:
        # returns a logger with the name of the module
        logger = get_logger(__name__)

        logger.info("This is a message")
        logger.debug("This is a debug message")
        logger.error("This is an error message")
        logger.warning("This is a warning message")

    :param name: the name of the desired logger
    :type name: str
    :param log_level: what level to set the logger to
    :type log_level: str
    :param fmt: the format of the log messages
    :type fmt: str
    :returns: logger instance
    :rtype: logging.Logger
    """
    # if name is None, then logger is the root logger
    # if not root logger, get the name of file without prefix.
    user_log_level = _get_log_level()
    if user_log_level != "developer":
        name = "SmartSim"

    logger = logging.getLogger(name)
    if log_level:
        logger.setLevel(log_level)
    else:
        log_level = user_log_level
    coloredlogs.install(level=log_level, logger=logger, fmt=fmt, stream=sys.stdout)
    return logger


def log_to_file(filename, log_level="debug"):
    """Installs a second filestream handler to the root logger,
    allowing subsequent logging calls to be sent to filename.

    :param filename: the name of the desired log file.
    :type filename: str

    :param log_level: as defined in get_logger.  Can be specified
                      to allow the file to store more or less verbose
                      logging information.
    :type log_level: int | str
    """
    logger = logging.getLogger("SmartSim")
    coloredlogs.install(stream=open(filename, "w+"), logger=logger, level=log_level)



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import abc

from ...error import AllocationError, LauncherError, SSUnsupportedError
from .stepInfo import UnmanagedStepInfo
from .stepMapping import StepMapping
from .taskManager import TaskManager


class Launcher(abc.ABC):  # pragma: no cover
    """Abstract base class of all launchers

    This class provides the interface between the experiment
    controller and the launcher layer. Each launcher supported
    in SmartSim should implement the methods in this class to
    be fully compatible.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def create_step(self, name, cwd, step_settings):
        raise NotImplementedError

    @abc.abstractmethod
    def get_step_update(self, step_names):
        raise NotImplementedError

    @abc.abstractmethod
    def get_step_nodes(self, step_names):
        raise NotImplementedError

    @abc.abstractmethod
    def run(self, step):
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self, step_name):
        raise NotImplementedError


class WLMLauncher(Launcher):  # cov-wlm
    """The base class for any Launcher that utilizes workload
    manager specific commands. This base class is used to provide
    implemented methods that are alike across all WLM launchers.
    """

    def __init__(self):
        super().__init__()
        self.task_manager = TaskManager()
        self.step_mapping = StepMapping()

    # every launcher utilizing this interface must have a map
    # of supported RunSettings types (see slurmLauncher.py for ex)
    def create_step(self, name, cwd, step_settings):  # cov-wlm
        """Create a WLM job step

        :param name: name of the entity to be launched
        :type name: str
        :param cwd: path to launch dir
        :type cwd: str
        :param step_settings: batch or run settings for entity
        :type step_settings: BatchSettings | RunSettings
        :raises SSUnsupportedError: if batch or run settings type isnt supported
        :raises LauncherError: if step creation fails
        :return: step instance
        :rtype: Step
        """
        try:
            settings_class = step_settings.__class__
            if settings_class in self.supported_rs:
                step_class = self.supported_rs[settings_class]
                step = step_class(name, cwd, step_settings)
                return step
            raise SSUnsupportedError(
                f"RunSettings type {type(step_settings)} not supported by this launcher"
            )
        except AllocationError as e:
            raise LauncherError("Step creation failed") from e

    # these methods are implemented in WLM launchers and
    # don't need to be covered here.

    def get_step_nodes(self, step_names):  # pragma: no cover
        raise SSUnsupportedError("Node acquisition not supported for this launcher")

    def run(self, step):  # pragma: no cover
        raise NotImplementedError

    def stop(self, step_name):  # pragma: no cover
        raise NotImplementedError

    def get_step_update(self, step_names):  # cov-wlm
        """Get update for a list of job steps

        :param step_names: list of job steps to get updates for
        :type step_names: list[str]
        :return: list of name, job update tuples
        :rtype: list[(str, StepInfo)]
        """
        updates = []

        # get updates of jobs managed by workload manager (PBS, Slurm, etc)
        # this is primarily batch jobs.
        s_names, step_ids = self.step_mapping.get_ids(step_names, managed=True)
        if len(step_ids) > 0:
            s_statuses = self._get_managed_step_update(step_ids)
            _updates = [(name, stat) for name, stat in zip(s_names, s_statuses)]
            updates.extend(_updates)

        # get updates of unmanaged jobs (Aprun, mpirun, etc)
        # usually jobs started and monitored through the Popen interface
        t_names, task_ids = self.step_mapping.get_ids(step_names, managed=False)
        if len(task_ids) > 0:
            t_statuses = self._get_unmanaged_step_update(task_ids)
            _updates = [(name, stat) for name, stat in zip(t_names, t_statuses)]
            updates.extend(_updates)

        return updates

    def _get_unmanaged_step_update(self, task_ids):  # cov-wlm
        """Get step updates for Popen managed jobs

        :param task_ids: task id to check
        :type task_ids: list[str]
        :return: list of step updates
        :rtype: list[StepInfo]
        """
        updates = []
        for task_id in task_ids:
            stat, rc, out, err = self.task_manager.get_task_update(task_id)
            update = UnmanagedStepInfo(stat, rc, out, err)
            updates.append(update)
        return updates

    def _get_managed_step_update(self, step_ids):  # pragma: no cover
        pass



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from shutil import which

"""
Parsers for various slurm functions.
"""


def parse_salloc(output):
    for line in output.split("\n"):
        if line.startswith("salloc: Granted job allocation"):
            return line.split()[-1]


def parse_salloc_error(output):
    """Parse and return error output of a failed salloc command

    :param output: stderr output of salloc command
    :type output: str
    :return: error message
    :rtype: str
    """
    salloc = which("salloc")
    # look for error first
    for line in output.split("\n"):
        if salloc and line.startswith(salloc + ": error:"):
            error = line.split("error:")[1]
            return error.strip()
        if line.startswith("salloc: error:"):
            error = line.split("error:")[1]
            return error.strip()
    # if no error line, take first line
    for line in output.split("\n"):
        if salloc and line.startswith(salloc + ": "):
            error = " ".join((line.split()[1:]))
            return error.strip()
        if line.startswith("salloc: "):
            error = " ".join((line.split()[1:]))
            return error.strip()
    # return None if we cant find error
    return None


def jobid_exact_match(parsed_id, job_id):
    """Check that job_id is an exact match and not
    the prefix of another job_id, like 1 and 11
    or 1.1 and 1.10. Works with job id or step
    id (i.e. with or without a '.' in the id)
    :param parsed_id: the id read from the line
    :type paserd_id: str
    :param job_id: the id to check for equality
    :type job_id: str
    """
    if "." in job_id:
        return parsed_id == job_id
    else:
        return parsed_id.split(".")[0] == job_id


def parse_sacct(output, job_id):
    """Parse and return output of the sacct command

    :param output: output of the sacct command
    :type output: str
    :param job_id: allocation id or job step id
    :type job_id: str
    :return: status and returncode
    :rtype: tuple
    """
    result = ("PENDING", None)
    for line in output.split("\n"):
        line = line.split("|")
        if len(line) >= 3:
            if jobid_exact_match(line[0], job_id):
                stat = line[1]
                code = line[2].split(":")[0]
                result = (stat, code)
                break
    return result


def parse_sstat_nodes(output, job_id):
    """Parse and return the sstat command

    This function parses and returns the nodes of
    a job in a list with the duplicates removed.

    :param output: output of the sstat command
    :type output: str
    :return: compute nodes of the allocation or job
    :rtype: list of str
    """
    nodes = []
    for line in output.split("\n"):
        sstat_string = line.split("|")

        # sometimes there are \n that we need to ignore
        if len(sstat_string) >= 2:
            if jobid_exact_match(sstat_string[0], job_id):
                node = sstat_string[1]
                nodes.append(node)
    return list(set(nodes))


def parse_step_id_from_sacct(output, step_name):
    """Parse and return the step id from a sacct command

    :param output: output of sacct --noheader -p
                   --format=jobname,jobid --job <alloc>
    :type output: str
    :param step_name: the name of the step to query
    :type step_name: str
    :return: the step_id
    :rtype: str
    """
    step_id = None
    for line in output.split("\n"):
        sacct_string = line.split("|")
        if len(sacct_string) >= 2:
            if sacct_string[0] == step_name:
                step_id = sacct_string[1]
    return step_id



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os.path as osp
import time

from ....log import get_logger
from ...utils.helpers import get_base_36_repr
from ..colocated import write_colocated_launch_script

logger = get_logger(__name__)


class Step:
    def __init__(self, name, cwd):
        self.name = self._create_unique_name(name)
        self.entity_name = name
        self.cwd = cwd
        self.managed = False

    def get_launch_cmd(self):
        raise NotImplementedError

    def _create_unique_name(self, entity_name):
        step_name = entity_name + "-" + get_base_36_repr(time.time_ns())
        return step_name

    def get_output_files(self):
        """Return two paths to error and output files based on cwd"""
        output = self.get_step_file(ending=".out")
        error = self.get_step_file(ending=".err")
        return output, error

    def get_step_file(self, ending=".sh", script_name=None):
        """Get the name for a file/script created by the step class

        Used for Batch scripts, mpmd scripts, etc"""
        if script_name:
            script_name = script_name if "." in script_name else script_name + ending
            return osp.join(self.cwd, script_name)
        return osp.join(self.cwd, self.entity_name + ending)

    def get_colocated_launch_script(self):
        # prep step for colocated launch if specifed in run settings
        script_path = self.get_step_file(script_name=".colocated_launcher.sh")

        db_settings = self.run_settings.colocated_db_settings

        # db log file causes write contention and kills performance so by
        # default we turn off logging unless user specified debug=True
        if db_settings.get("debug", False):
            db_log_file = self.get_step_file(ending="-db.log")
        else:
            db_log_file = "/dev/null"

        # if user specified to use taskset with local launcher
        # (not allowed b/c MacOS doesn't support it)
        # TODO: support this only on linux
        if (
            self.__class__.__name__ == "LocalStep"
            and db_settings["limit_app_cpus"] is True
        ):  # pragma: no cover
            logger.warning("Setting limit_app_cpus=False for local launcher")
            db_settings["limit_app_cpus"] = False

        # write the colocated wrapper shell script to the directory for this
        # entity currently being prepped to launch
        write_colocated_launch_script(script_path, db_log_file, db_settings)
        return script_path



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from shutil import which

from .._core.launcher.slurm.slurmCommands import salloc, scancel, scontrol, sinfo
from .._core.launcher.slurm.slurmParser import parse_salloc, parse_salloc_error
from .._core.launcher.util.launcherUtil import ComputeNode, Partition
from .._core.utils.helpers import init_default
from ..error import AllocationError, LauncherError, SmartSimError
from ..log import get_logger

logger = get_logger(__name__)


def get_allocation(nodes=1, time=None, account=None, options=None):
    """Request an allocation

    This function requests an allocation with the specified arguments.
    Anything passed to the options will be processed as a Slurm
    argument and appended to the salloc command with the appropriate
    prefix (e.g. "-" or "--").

    The options can be used to pass extra settings to the
    workload manager such as the following for Slurm:

        - nodelist="nid00004"

    For arguments without a value, pass None or and empty
    string as the value. For Slurm:

        - exclusive=None

    :param nodes: number of nodes for the allocation, defaults to 1
    :type nodes: int, optional
    :param time: wall time of the allocation, HH:MM:SS format, defaults to None
    :type time: str, optional
    :param account: account id for allocation, defaults to None
    :type account: str, optional
    :param options: additional options for the slurm wlm, defaults to None
    :type options: dict[str, str], optional
    :raises LauncherError: if the allocation is not successful
    :return: the id of the allocation
    :rtype: str
    """
    if not which("salloc"):
        raise LauncherError(
            "Attempted slurm function without access to slurm(salloc) at the call site"
        )

    options = init_default({}, options, dict)

    salloc_args = _get_alloc_cmd(nodes, time, account, options=options)
    debug_msg = " ".join(salloc_args[1:])
    logger.debug(f"Allocation settings: {debug_msg}")

    _, err = salloc(salloc_args)
    alloc_id = parse_salloc(err)
    if alloc_id:
        logger.info(f"Allocation successful with Job ID: {str(alloc_id)}")
    else:
        logger.debug(err)
        error = parse_salloc_error(err)
        if not error:
            logger.error(err)
            raise AllocationError("Error retrieving Slurm allocation")
        raise AllocationError(error)
    return str(alloc_id)


def release_allocation(alloc_id):
    """Free an allocation's resources

    :param alloc_id: allocation id
    :type alloc_id: str
    :raises LauncherError: if allocation could not be freed
    """
    if not which("scancel"):
        raise LauncherError(
            "Attempted slurm function without access to slurm(salloc) at the call site"
        )

    logger.info(f"Releasing allocation: {alloc_id}")
    returncode, _, _ = scancel([str(alloc_id)])

    if returncode != 0:
        logger.error(f"Unable to revoke your allocation for jobid {str(alloc_id)}")
        logger.error(
            "The job may have already timed out, or you may need to cancel the job manually"
        )
        raise AllocationError(
            f"Unable to revoke your allocation for jobid  {str(alloc_id)}"
        )

    logger.info(f"Successfully freed allocation {alloc_id}")


def validate(nodes=1, ppn=1, partition=None):
    """Check that there are sufficient resources in the provided Slurm partitions.

    if no partition is provided, the default partition is found and used.

    :param nodes: Override the default node count to validate, defaults to 1
    :type nodes: int, optional
    :param ppn: Override the default processes per node to validate, defaults to 1
    :type ppn: int, optional
    :param partition: partition to validate, defaults to None
    :type partition: str, optional
    :raises: LauncherError
    :returns: True if resources are available, False otherwise
    :rtype: bool
    """
    sys_partitions = _get_system_partition_info()

    n_avail_nodes = 0
    avail_nodes = set()

    p_name = partition
    if p_name is None or p_name == "default":
        try:
            p_name = get_default_partition()
        except LauncherError as e:
            raise LauncherError(
                "No partition provided and default partition could not be found"
            ) from e

    if not p_name in sys_partitions:
        raise LauncherError(f"Partition {p_name} is not found on this system")

    for node in sys_partitions[p_name].nodes:
        if node.ppn >= ppn:
            avail_nodes.add(node)

    n_avail_nodes = len(avail_nodes)
    logger.debug(f"Found {n_avail_nodes} nodes that match the constraints provided")

    if n_avail_nodes < nodes:
        logger.warning(
            f"{nodes} nodes are not available on the specified partitions.  Only "
            f"{n_avail_nodes} nodes available."
        )
        return False

    logger.info("Successfully validated Slurm with sufficient resources")
    return True


def get_default_partition():
    """Returns the default partition from Slurm

    This default partition is assumed to be the partition with
    a star following its partition name in sinfo output

    :returns: the name of the default partition
    :rtype: str
    """
    sinfo_output, _ = sinfo(["--noheader", "--format", "%P"])

    default = None
    for line in sinfo_output.split("\n"):
        if line.endswith("*"):
            default = line.strip("*")

    if not default:
        raise LauncherError("Could not find default partition!")
    return default


def _get_system_partition_info():
    """Build a dictionary of slurm partitions
    :returns: dict of Partition objects
    :rtype: dict
    """

    sinfo_output, _ = sinfo(["--noheader", "--format", "%R %n %c"])

    partitions = {}
    for line in sinfo_output.split("\n"):
        line = line.strip()
        if line == "":
            continue

        p_info = line.split(" ")
        p_name = p_info[0]
        p_node = p_info[1]
        p_ppn = int(p_info[2])

        if not p_name in partitions:
            partitions.update({p_name: Partition()})

        partitions[p_name].name = p_name
        partitions[p_name].nodes.add(ComputeNode(node_name=p_node, node_ppn=p_ppn))

    return partitions


def _get_alloc_cmd(nodes, time, account, options=None):
    """Return the command to request an allocation from Slurm with
    the class variables as the slurm options."""

    salloc_args = [
        "--no-shell",
        "-N",
        str(nodes),
        "-J",
        "SmartSim",
    ]
    # TODO check format here
    if time:
        salloc_args.extend(["-t", time])
    if account:
        salloc_args.extend(["-A", str(account)])

    for opt, val in options.items():
        if opt not in ["t", "time", "N", "nodes", "A", "account"]:
            short_arg = bool(len(str(opt)) == 1)
            prefix = "-" if short_arg else "--"
            if not val:
                salloc_args += [prefix + opt]
            else:
                if short_arg:
                    salloc_args += [prefix + opt, str(val)]
                else:
                    salloc_args += ["=".join((prefix + opt, str(val)))]

    return salloc_args


def get_hosts():
    """Get the name of the nodes used in a slurm allocation.

    .. note::

        This method requires access to ``scontrol`` from the node
        on which it is run

    :returns: Names of the host nodes
    :rtype: list[str]
    :raises LauncherError: Could not access ``scontrol``
    :raises SmartSimError: ``SLURM_JOB_NODELIST`` is not set
    """
    if "SLURM_JOB_NODELIST" in os.environ:
        if not which("scontrol"):
            raise LauncherError(
                (
                    "Attempted slurm function without access to "
                    "slurm(scontrol) at the call site"
                )
            )
        nodelist, _ = scontrol(
            ["show", "hostnames", os.environ.get("SLURM_JOB_NODELIST")]
        )
        return sorted(nodelist.split())
    raise SmartSimError("Could not parse allocation nodes from SLURM_JOB_NODELIST")


def get_queue():
    """Get the name of queue in a slurm allocation.

    :returns: The name of the queue
    :rtype: str
    :raises SmartSimError: ``SLURM_JOB_PARTITION`` is not set
    """
    if "SLURM_JOB_PARTITION" in os.environ:
        return os.environ.get("SLURM_JOB_PARTITION")
    raise SmartSimError("Could not parse queue from SLURM_JOB_PARTITION")


def get_tasks():
    """Get the number of tasks in a slurm allocation.

    :returns: Then number of tasks in the allocation
    :rtype: int
    :raises SmartSimError: ``SLURM_NTASKS`` is not set
    """
    if "SLURM_NTASKS" in os.environ:
        return int(os.environ.get("SLURM_NTASKS"))
    raise SmartSimError("Could not parse number of requested tasks from SLURM_NTASKS")


def get_tasks_per_node():
    """Get the number of tasks per each node in a slurm allocation.

    .. note::

        This method requires access to ``scontrol`` from the node
        on which it is run

    :returns: Map of nodes to number of tasks on that node
    :rtype: dict[str, int]
    :raises SmartSimError: ``SLURM_TASKS_PER_NODE`` is not set
    """
    if "SLURM_TASKS_PER_NODE" in os.environ:
        tasks_per_node_strs = os.environ.get("SLURM_TASKS_PER_NODE").split(",")
        tasks_per_node_list = []
        for tasks_per_node_str in tasks_per_node_strs:
            if "(" in tasks_per_node_str:
                tasks, quantity = tasks_per_node_str.split("(")
                quantity = quantity.rstrip(")").lstrip("x")
                tasks_per_node_list.extend([int(tasks)] * int(quantity))
            else:
                tasks_per_node_list.append(int(tasks_per_node_str))
        return dict(zip(get_hosts(), tasks_per_node_list))
    raise SmartSimError("Could not parse tasks per node from SLURM_TASKS_PER_NODE")



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import sys
from pathlib import Path

import pkg_resources
from tabulate import tabulate

from smartsim._core._cli.utils import color_bool, pip_install
from smartsim._core._install import builder
from smartsim._core._install.buildenv import BuildEnv, SetupError, Version_, Versioner
from smartsim._core._install.builder import BuildError
from smartsim._core.config import CONFIG
from smartsim._core.utils.helpers import installed_redisai_backends
from smartsim.error import SSConfigError
from smartsim.log import get_logger

smart_logger_format = "[%(name)s] %(levelname)s %(message)s"
logger = get_logger("Smart", fmt=smart_logger_format)

# NOTE: all smartsim modules need full paths as the smart cli
#       may be installed into a different directory.


def _install_torch_from_pip(versions, device="cpu", verbose=False):

    packages = []
    end_point = None

    if sys.platform == "darwin":
        if device == "gpu":
            logger.warning("GPU support is not available on Mac OS X")
        # The following is deliberately left blank as there is no
        # alternative package available on Mac OS X
        device_suffix = ""
        end_point = None

    # if we are on linux cpu, either CUDA or CPU must be installed
    elif sys.platform == "linux":
        end_point = "https://download.pytorch.org/whl/torch_stable.html"
        if device in ["gpu", "cuda"]:
            device_suffix = versions.TORCH_CUDA_SUFFIX
        elif device == "cpu":
            device_suffix = versions.TORCH_CPU_SUFFIX

    packages.append(f"torch=={versions.TORCH}{device_suffix}")
    packages.append(f"torchvision=={versions.TORCHVISION}{device_suffix}")

    pip_install(packages, end_point=end_point, verbose=verbose)


class Build:
    def __init__(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-v",
            action="store_true",
            default=False,
            help="Enable verbose build process",
        )
        parser.add_argument(
            "--device",
            type=str.lower,
            default="cpu",
            help="Device to build ML runtimes for (cpu || gpu)",
        )
        parser.add_argument(
            "--no_pt",
            action="store_true",
            default=False,
            help="Do not build PyTorch backend",
        )
        parser.add_argument(
            "--no_tf",
            action="store_true",
            default=False,
            help="Do not build TensorFlow backend",
        )
        parser.add_argument(
            "--onnx",
            action="store_true",
            default=False,
            help="Build ONNX backend (off by default)",
        )
        parser.add_argument(
            "--torch_dir",
            default=None,
            type=str,
            help="Path to custom <path>/torch/share/cmake/Torch/ directory (ONLY USE IF NEEDED)",
        )
        parser.add_argument(
            "--libtensorflow_dir",
            default=None,
            type=str,
            help="Path to custom libtensorflow directory (ONLY USED IF NEEDED)",
        )
        parser.add_argument(
            "--only_python_packages",
            action="store_true",
            default=False,
            help="If true, only install the python packages (i.e. skip backend builds)",
        )
        parser.add_argument(
            "--keydb",
            action="store_true",
            default=False,
            help="Build KeyDB instead of Redis",
        )
        args = parser.parse_args(sys.argv[2:])
        self.verbose = args.v
        self.keydb = args.keydb

        # torch and tf build by default
        pt = not args.no_pt
        tf = not args.no_tf
        onnx = args.onnx

        logger.info("Running SmartSim build process...")
        try:
            logger.info("Checking requested versions...")
            self.versions = Versioner()

            if args.only_python_packages:
                logger.info("Only installing Python packages...skipping build")
                self.build_env = BuildEnv(checks=False)
                if not args.no_pt:
                    self.install_torch(device=args.device)
            else:
                logger.info("Checking for build tools...")
                self.build_env = BuildEnv()

                if self.verbose:
                    logger.info("Build Environment:")
                    env = self.build_env.as_dict()
                    print(tabulate(env, headers=env.keys(), tablefmt="github"), "\n")
                if self.keydb:
                    self.versions.REDIS = Version_("6.2.0")
                    self.versions.REDIS_URL = "https://github.com/EQ-Alpha/KeyDB"
                    self.versions.REDIS_BRANCH = "v6.2.0"
                    CONFIG.conf_path = Path(CONFIG.core_path, "config", "keydb.conf")
                    if not CONFIG.conf_path.resolve().is_file():
                        raise SSConfigError(
                            "Database configuration file at REDIS_CONF could not be found"
                        )

                if self.verbose:
                    db_name = "KEYDB" if self.keydb else "REDIS"
                    logger.info("Version Information:")
                    vers = self.versions.as_dict(db_name=db_name)
                    print(tabulate(vers, headers=vers.keys(), tablefmt="github"), "\n")

                # REDIS/KeyDB
                self.build_database()

                # REDISAI
                self.build_redis_ai(
                    str(args.device),
                    pt,
                    tf,
                    onnx,
                    args.torch_dir,
                    args.libtensorflow_dir,
                )

                backends = [
                    backend.capitalize() for backend in installed_redisai_backends()
                ]
                logger.info(
                    (", ".join(backends) if backends else "No") + " backend(s) built"
                )

        except (SetupError, BuildError) as e:
            logger.error(str(e))
            sys.exit(1)

        logger.info("SmartSim build complete!")

    def build_database(self):
        # check database installation
        database_name = "KeyDB" if self.keydb else "Redis"
        database_builder = builder.DatabaseBuilder(
            self.build_env(), self.build_env.MALLOC, self.build_env.JOBS, self.verbose
        )
        if not database_builder.is_built:
            logger.info(
                f"Building {database_name} version {self.versions.REDIS} from {self.versions.REDIS_URL}"
            )
            database_builder.build_from_git(
                self.versions.REDIS_URL, self.versions.REDIS_BRANCH
            )
            database_builder.cleanup()
        logger.info(f"{database_name} build complete!")

    def build_redis_ai(
        self, device, torch=True, tf=True, onnx=False, torch_dir=None, libtf_dir=None
    ):

        # make sure user isn't trying to do something silly on MacOS
        if self.build_env.PLATFORM == "darwin" and device == "gpu":
            raise BuildError("SmartSim does not support GPU on MacOS")

        # decide which runtimes to build
        print("\nML Backends Requested")
        backends_table = [
            ["PyTorch", self.versions.TORCH, color_bool(torch)],
            ["TensorFlow", self.versions.TENSORFLOW, color_bool(tf)],
            ["ONNX", self.versions.ONNX or "Unavailable", color_bool(onnx)],
        ]
        print(tabulate(backends_table, tablefmt="fancy_outline"), end="\n\n")
        print(f"Building for GPU support: {color_bool(device == 'gpu')}\n")

        self.check_backends_install()

        # Check for onnx and tf in user python environemnt and prompt user
        # to download them if they are not installed. this should not break
        # the build however, as we use onnx and tf directly from RAI instead
        # of pip like we do PyTorch.
        if onnx:
            self.check_onnx_install()
        if tf:
            self.check_tf_install()

        # TORCH
        if torch:
            if torch_dir:
                torch_dir = Path(torch_dir).resolve()
                if not torch_dir.is_dir():
                    # we will always be able to find a torch version downloaded by
                    # pip so if we can't find it we know the user suggested a torch
                    # installation path that doesn't exist
                    logger.error("Could not find requested user Torch installation")
                    sys.exit(1)
            else:
                # install pytorch wheel, and get the path to the cmake dir
                # we will use in the RAI build
                self.install_torch(device=device)
                torch_dir = self.build_env.torch_cmake_path

        if tf:
            if libtf_dir:
                libtf_dir = Path(libtf_dir).resolve()

        rai_builder = builder.RedisAIBuilder(
            build_env=self.build_env(),
            torch_dir=str(torch_dir) if torch_dir else None,
            libtf_dir=str(libtf_dir) if libtf_dir else None,
            build_torch=torch,
            build_tf=tf,
            build_onnx=onnx,
            jobs=self.build_env.JOBS,
            verbose=self.verbose,
        )

        if rai_builder.is_built:
            logger.info("RedisAI installed. Run `smart clean` to remove.")
        else:
            # get the build environment, update with CUDNN env vars
            # if present and building for GPU, otherwise warn the user
            build_env = self.build_env()
            if device == "gpu":
                gpu_env = self.build_env.get_cudnn_env()
                cudnn_env_vars = [
                    "CUDNN_LIBRARY",
                    "CUDNN_INCLUDE_DIR",
                    "CUDNN_INCLUDE_PATH",
                    "CUDNN_LIBRARY_PATH",
                ]
                if not gpu_env:
                    logger.warning(
                        f"CUDNN environment variables not found.\n"
                        + f"Looked for {cudnn_env_vars}"
                    )
                else:
                    build_env.update(gpu_env)
            # update RAI build env with cudnn env vars
            rai_builder.env = build_env

            logger.info(
                f"Building RedisAI version {self.versions.REDISAI}"
                f" from {self.versions.REDISAI_URL}"
            )

            # NOTE: have the option to add other builds here in the future
            # like "from_tarball"
            rai_builder.build_from_git(
                self.versions.REDISAI_URL, self.versions.REDISAI_BRANCH, device
            )
            logger.info("ML Backends and RedisAI build complete!")

    def infer_torch_device(self):
        backend_torch_path = f"{CONFIG.lib_path}/backends/redisai_torch"
        device = "cpu"
        if Path(f"{backend_torch_path}/lib/libtorch_cuda.so").is_file():
            device = "gpu"
        return device

    def install_torch(self, device="cpu"):
        """Torch shared libraries installed by pip are used in the build
        for SmartSim backends so we download them here.
        """

        if not self.build_env.check_installed("torch", self.versions.TORCH):
            inferred_device = self.infer_torch_device()
            if (inferred_device == "gpu") and (device == "cpu"):
                logger.warning("CPU requested, but GPU backend is available")
            _install_torch_from_pip(self.versions, device, self.verbose)
        # if torch already installed, check the versions to make sure correct
        # torch version is downloaded for that particular device
        else:
            installed = Version_(pkg_resources.get_distribution("torch").version)
            if device == "gpu":
                # if torch version is x.x.x+cpu
                if "cpu" in installed.patch:
                    msg = (
                        "Torch CPU is currently installed but torch GPU requested. Uninstall all torch packages "
                        + "and run the `smart build` command again to obtain Torch GPU libraries"
                    )
                    logger.warning(msg)

            if device == "cpu":
                # if torch version if x.x.x then we need to install the cpu version
                if "cpu" not in installed.patch and not self.build_env.is_macos():
                    msg = (
                        "Torch GPU installed in python environment but requested Torch CPU. "
                        + " Run `pip uninstall torch torchvision` and run `smart build` again"
                    )
                    logger.error(msg)  # error because this is usually fatal
            logger.info(f"Torch {self.versions.TORCH} installed in Python environment")

    def check_onnx_install(self):
        """Check Python environment for ONNX installation"""
        if not self.versions.ONNX:
            py_version = sys.version_info
            msg = (
                "An onnx wheel is not available for "
                f"Python {py_version.major}.{py_version.minor}. "
                "Instead consider using Python 3.8 or 3.9 with Onnx "
            )
            if sys.platform == "linux":
                msg += "1.2.5 or "
            msg += "1.2.7."
            raise SetupError(msg)
        try:
            if not self.build_env.check_installed("onnx", self.versions.ONNX):
                msg = (
                    f"ONNX {self.versions.ONNX} not installed in python environment. "
                    + f"Consider installing onnx=={self.versions.ONNX} with pip"
                )
                logger.warning(msg)
            else:
                logger.info(
                    f"ONNX {self.versions.ONNX} installed in Python environment"
                )
        except SetupError as e:
            logger.warning(str(e))

    def check_tf_install(self):
        """Check Python environment for TensorFlow installation"""

        try:
            if not self.build_env.check_installed(
                "tensorflow", self.versions.TENSORFLOW
            ):
                msg = (
                    f"TensorFlow {self.versions.TENSORFLOW} not installed in Python environment. "
                    + f"Consider installing tensorflow=={self.versions.TENSORFLOW} with pip"
                )
                logger.warning(msg)
            else:
                logger.info(
                    f"TensorFlow {self.versions.TENSORFLOW} installed in Python environment"
                )
        except SetupError as e:
            logger.warning(str(e))

    def check_backends_install(self):
        """Checks if backends have already been installed.
        Logs details on how to proceed forward
        if the RAI_PATH environment variable is set or if
        backends have already been installed.
        """
        if os.environ.get("RAI_PATH"):
            if installed_redisai_backends():
                logger.error(
                    f"There is no need to build. Backends are already built and specified in the environment at 'RAI_PATH': {CONFIG.redisai}"
                )
            else:
                logger.error(
                    f"Before running 'smart build', unset your RAI_PATH environment variable with 'unset RAI_PATH'."
                )
            sys.exit(1)
        else:
            if installed_redisai_backends():
                logger.error(
                    "If you wish to re-run `smart build`, you must first run `smart clean`."
                )
                sys.exit(1)



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time

from ...status import STATUS_NEW


class Job:
    """Keep track of various information for the controller.
    In doing so, continuously add various fields of information
    that is queryable by the user through interface methods in
    the controller class.
    """

    def __init__(self, job_name, job_id, entity, launcher, is_task):
        """Initialize a Job.

        :param job_name: Name of the job step
        :type job_name: str
        :param job_id: The id associated with the job
        :type job_id: str
        :param entity: The SmartSim entity associated with the job
        :type entity: SmartSimEntity
        :param launcher: Launcher job was started with
        :type launcher: str
        :param is_task: process monitored by TaskManager (True) or the WLM (True)
        :type is_task: bool
        """
        self.name = job_name
        self.jid = job_id
        self.entity = entity
        self.status = STATUS_NEW
        self.raw_status = None  # status before smartsim status mapping is applied
        self.returncode = None
        self.output = None  # only populated if it's system related (e.g. a command failed immediately)
        self.error = None  # same as output
        self.hosts = []  # currently only used for DB jobs
        self.launched_with = launcher
        self.is_task = is_task
        self.start_time = time.time()
        self.history = History()

    @property
    def ename(self):
        """Return the name of the entity this job was created from"""
        return self.entity.name

    def set_status(self, new_status, raw_status, returncode, error=None, output=None):
        """Set the status  of a job.

        :param new_status: The new status of the job
        :type new_status: str
        :param returncode: The return code for the job
        :type return_code: str
        """
        self.status = new_status
        self.raw_status = raw_status
        self.returncode = returncode
        self.error = error
        self.output = output

    def record_history(self):
        """Record the launching history of a job."""
        job_time = time.time() - self.start_time
        self.history.record(self.jid, self.status, self.returncode, job_time)

    def reset(self, new_job_name, new_job_id, is_task):
        """Reset the job in order to be able to restart it.

        :param new_job_name: name of the new job step
        :type new_job_name: str
        :param new_job_id: new job id to launch under
        :type new_job_id: str
        :param is_task: process monitored by TaskManager (True) or the WLM (True)
        :type is_task: bool
        """
        self.name = new_job_name
        self.jid = new_job_id
        self.status = STATUS_NEW
        self.returncode = None
        self.output = None
        self.error = None
        self.hosts = []
        self.is_task = is_task
        self.start_time = time.time()
        self.history.new_run()

    def error_report(self):
        """A descriptive error report based on job fields

        :return: error report for display in terminal
        :rtype: str
        """
        warning = f"{self.ename} failed. See below for details \n"
        if self.error:
            warning += (
                f"{self.entity.type} {self.ename} produced the following error \n"
            )
            warning += f"Error: {self.error} \n"
        if self.output:
            warning += f"Output: {self.output} \n"
        warning += f"Job status at failure: {self.status} \n"
        warning += f"Launcher status at failure: {self.raw_status} \n"
        warning += f"Job returncode: {self.returncode} \n"
        warning += f"Error and output file located at: {self.entity.path}"
        return warning

    def __str__(self):
        """Return user-readable string of the Job

        :returns: A user-readable string of the Job
        :rtype: str
        """
        if self.jid:
            job = "{}({}): {}"
            return job.format(self.ename, self.jid, self.status)
        else:
            job = "{}: {}"
            return job.format(self.ename, self.status)


class History:
    """History of a job instance. Holds various attributes based
    on the previous launches of a job.
    """

    def __init__(self, runs=0):
        """Init a history object for a job

        :param runs: number of runs so far, defaults to 0
        :type runs: int, optional
        """
        self.runs = runs
        self.jids = dict()
        self.statuses = dict()
        self.returns = dict()
        self.job_times = dict()

    def record(self, job_id, status, returncode, job_time):
        """record the history of a job"""
        self.jids[self.runs] = job_id
        self.statuses[self.runs] = status
        self.returns[self.runs] = returncode
        self.job_times[self.runs] = job_time

    def new_run(self):
        """increment run total"""
        self.runs += 1



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from functools import lru_cache
from pathlib import Path

import psutil

from ...error import SSConfigError
from ..utils.helpers import expand_exe_path

# Configuration Values
#
# These values can be set through environment variables to
# override the default behavior of SmartSim.
#
# RAI_PATH
#   - Path to the RAI shared library
#   - Default: /smartsim/smartsim/_core/lib/redisai.so
#
# REDIS_CONF
#   - Path to the redis.conf file
#   - Default: /SmartSim/smartsim/config/redis6.conf
#
# REDIS_PATH
#   - Path to the redis-server executable
#   - Default: /SmartSim/smartsim/bin/redis-server
#
# REDIS_CLI_PATH
#   - Path to the redis-cli executable
#   - Default: /SmartSim/smartsim/bin/redis-cli
#
# SMARTSIM_LOG_LEVEL
#   - Log level for SmartSim
#   - Default: info
#
# SMARTSIM_JM_INTERVAL
#   - polling interval for communication with scheduler
#   - default: 10 seconds
#


# Testing Configuration Values
#
# SMARTSIM_TEST_INTERFACE
#  - Network interface to use for testing
#  - Default: auto-detected
#
# SMARTSIM_TEST_LAUNCHER
#  - type of launcher to use for testing
#  - Default: Local
#
# SMARTSIM_TEST_DEVICE
#  - CPU or GPU for model serving tests
#  - Default: CPU
#
# SMARTSIM_TEST_ACCOUNT
#  - Account used to run full launcher test suite on external systems
#  - Default: None


class Config:
    def __init__(self):
        # SmartSim/smartsim/_core
        self.core_path = Path(os.path.abspath(__file__)).parent.parent

        dependency_path = os.environ.get("SMARTSIM_DEP_INSTALL_PATH", self.core_path)

        self.lib_path = Path(dependency_path, "lib").resolve()
        self.bin_path = Path(dependency_path, "bin").resolve()
        self.conf_path = Path(dependency_path, "config", "redis6.conf")

    @property
    def redisai(self) -> str:
        rai_path = self.lib_path / "redisai.so"
        redisai = Path(os.environ.get("RAI_PATH", rai_path)).resolve()
        if not redisai.is_file():
            raise SSConfigError(
                "RedisAI dependency not found. Build with `smart` cli or specify RAI_PATH"
            )
        return str(redisai)

    @property
    def database_conf(self) -> str:
        conf = Path(os.environ.get("REDIS_CONF", self.conf_path)).resolve()
        if not conf.is_file():
            raise SSConfigError(
                "Database configuration file at REDIS_CONF could not be found"
            )
        return str(conf)

    @property
    def database_exe(self) -> str:
        try:
            database_exe = next(self.bin_path.glob("*-server"))
            database = Path(os.environ.get("REDIS_PATH", database_exe)).resolve()
            exe = expand_exe_path(str(database))
            return exe
        except (TypeError, FileNotFoundError) as e:
            raise SSConfigError(
                "Specified database binary at REDIS_PATH could not be used"
            ) from e

    @property
    def database_cli(self) -> str:
        try:
            redis_cli_exe = next(self.bin_path.glob("*-cli"))
            redis_cli = Path(os.environ.get("REDIS_CLI_PATH", redis_cli_exe)).resolve()
            exe = expand_exe_path(str(redis_cli))
            return exe
        except (TypeError, FileNotFoundError) as e:
            raise SSConfigError(
                "Specified Redis binary at REDIS_CLI_PATH could not be used"
            ) from e

    @property
    def log_level(self) -> str:
        return os.environ.get("SMARTSIM_LOG_LEVEL", "info")

    @property
    def jm_interval(self) -> int:
        return int(os.environ.get("SMARTSIM_JM_INTERVAL", 10))

    @property
    def wlm_trials(self) -> int:
        return int(os.environ.get("SMARTSIM_WLM_TRIALS", 10))

    @property
    def test_launcher(self) -> str:
        return os.environ.get("SMARTSIM_TEST_LAUNCHER", "local")

    @property
    def test_device(self) -> str:
        return os.environ.get("SMARTSIM_TEST_DEVICE", "CPU")

    @property
    def test_port(self) -> int:
        return int(os.environ.get("SMARTSIM_TEST_PORT", 6780))

    @property
    def test_interface(self) -> str:
        interface = os.environ.get("SMARTSIM_TEST_INTERFACE", None)
        if not interface:
            # try to pick a sensible one
            net_if_addrs = psutil.net_if_addrs()
            if "ipogif0" in net_if_addrs:
                return "ipogif0"
            elif "ib0" in net_if_addrs:
                return "ib0"
            # default to aries network
            return "ipogif0"
        else:
            return interface

    @property
    def test_account(self) -> str:
        # no account by default
        return os.environ.get("SMARTSIM_TEST_ACCOUNT", "")


@lru_cache(maxsize=128, typed=False)
def get_config():

    # wrap into a function with a cached result
    return Config()



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import os
from shutil import which

from smartsim.error.errors import LauncherError, SmartSimError

from .._core.launcher.pbs.pbsCommands import qstat


def get_hosts():
    """Get the name of the hosts used in a PBS allocation.

    :returns: Names of the host nodes
    :rtype: list[str]
    :raises SmartSimError: ``PBS_NODEFILE`` is not set
    """
    hosts = []
    if "PBS_NODEFILE" in os.environ:
        node_file = os.environ["PBS_NODEFILE"]
        with open(node_file, "r") as f:
            for line in f.readlines():
                host = line.split(".")[0].strip()
                hosts.append(host)
        # account for repeats in PBS_NODEFILE
        return sorted(list(set(hosts)))
    raise SmartSimError(
        "Could not parse interactive allocation nodes from PBS_NODEFILE"
    )


def get_queue():
    """Get the name of queue in a PBS allocation.

    :returns: The name of the queue
    :rtype: str
    :raises SmartSimError: ``PBS_QUEUE`` is not set
    """
    if "PBS_QUEUE" in os.environ:
        return os.environ.get("PBS_QUEUE")
    raise SmartSimError("Could not parse queue from SLURM_JOB_PARTITION")


def get_tasks():
    """Get the number of processes on each chunk in a PBS allocation.

    .. note::

        This method requires ``qstat`` be installed on the
        node from which it is run.

    :returns: Then number of tasks in the allocation
    :rtype: int
    :raises LauncherError: Could not access ``qstat``
    :raises SmartSimError: ``PBS_JOBID`` is not set
    """
    if "PBS_JOBID" in os.environ:
        if not which("qstat"):
            raise LauncherError(
                (
                    "Attempted PBS function without access to "
                    "PBS(qstat) at the call site"
                )
            )
        job_id = os.environ.get("PBS_JOBID")
        job_info_str, _ = qstat(["-f", "-F", "json", job_id])
        job_info = json.loads(job_info_str)
        return int(job_info["Jobs"][job_id]["resources_used"]["ncpus"])
    raise SmartSimError(
        "Could not parse number of requested tasks without an allocation"
    )


def get_tasks_per_node():
    """Get the number of processes on each chunk in a PBS allocation.

    .. note::

        This method requires ``qstat`` be installed on the
        node from which it is run.

    :returns: Map of chunks to number of processes on that chunck
    :rtype: dict[str, int]
    :raises LauncherError: Could not access ``qstat``
    :raises SmartSimError: ``PBS_JOBID`` is not set
    """
    if "PBS_JOBID" in os.environ:
        if not which("qstat"):
            raise LauncherError(
                (
                    "Attempted PBS function without access to "
                    "PBS(qstat) at the call site"
                )
            )
        job_id = os.environ.get("PBS_JOBID")
        job_info_str, _ = qstat(["-f", "-F", "json", job_id])
        job_info = json.loads(job_info_str)
        chunks_and_ncpus = job_info["Jobs"][job_id]["exec_vnode"]  # type: str

        chunk_cpu_map = {}
        for cunck_and_ncpu in chunks_and_ncpus.split("+"):
            chunk, ncpu = cunck_and_ncpu.strip("()").split(":")
            ncpu = ncpu.lstrip("ncpus=")
            chunk_cpu_map[chunk] = int(ncpu)

        return chunk_cpu_map
    raise SmartSimError("Could not parse tasks per node without an allocation")



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import subprocess
from pathlib import Path

from smartsim._core._install.buildenv import SetupError
from smartsim._core._install.builder import BuildError
from smartsim._core.utils import colorize
from smartsim.log import get_logger

smart_logger_format = "[%(name)s] %(levelname)s %(message)s"
logger = get_logger("Smart", fmt=smart_logger_format)


def get_install_path():
    try:
        import smartsim as _
    except (ImportError, ModuleNotFoundError):
        raise SetupError("Could not import SmartSim") from None

    # find the path to the setup script
    package_path = Path(_.__path__[0]).resolve()
    if not package_path.is_dir():
        raise SetupError("Could not find SmartSim installation site")

    return package_path


def color_bool(trigger=True):
    _color = "green" if trigger else "red"
    return colorize(str(trigger), color=_color)


def pip_install(packages, end_point=None, verbose=False):
    """Install a pip package to be used in the SmartSim build
    Currently only Torch shared libraries are re-used for the build
    """
    if end_point:
        packages.append(f"-f {end_point}")
    packages = " ".join(packages)

    # form pip install command
    cmd = ["python", "-m", "pip", "install", packages]
    cmd = " ".join(cmd)

    if verbose:
        logger.info(f"Installing packages {packages}...")
    proc = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    _, err = proc.communicate()
    returncode = int(proc.returncode)
    if returncode != 0:
        error = f"{packages} installation failed with exitcode {returncode}\n"
        error += err.decode("utf-8")
        raise BuildError(error)
    if verbose:
        logger.info(f"{packages} installed successfully")



# BSD 2-Clause License
#
# Copyright (c) 2021-2023, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Welcome to the SmartSim setup.py
#
# The following environment variables represent build time
# options for SmartSim. These are only relevant to when a user
# or CI is invoking the setup.py script.
#
#
# NO_CHECKS
#   If set to 1, the build process will not check for
#   build dependencies like make, gcc, etc
#
# SMARTSIM_REDIS
#   The version of redis to retrive and build with
#
# SMARTSIM_REDIS_URL
#   The URL from which to retrieve redis source code
#
# SMARTREDIS_VERSION
#   The version of SmartRedis to install.
#
# CC
#   The C compiler to use
#
# CXX
#   the CPP compiler to use
#
# MALLOC
#   The memory allocator to use for Redis (options: libc, jemalloc)
#
# BUILD_JOBS
#   Number of jobs to use in the build (defaults to max on node)
#
# SMARTSIM_SUFFIX
#  if set, the version number is set to a developer build
#  with the current version, git-sha, and suffix. This version
#  is then written into SmartSim/smartsim/version.py
#
#
# This future is needed to print Python2 EOL message
from __future__ import print_function
import sys

if sys.version_info < (3,):
    print("Python 2 has reached end-of-life and is not supported by SmartSim")
    sys.exit(-1)


import os
import importlib.util
from pathlib import Path

from setuptools import setup
from setuptools.dist import Distribution
from setuptools.command.install import install
from setuptools.command.build_py import build_py

# Some necessary evils we have to do to be able to use
# the _install tools in smartsim/smartsim/_core/_install
# in both the setup.py and in the smart cli

# import the installer classes
setup_path = Path(os.path.abspath(os.path.dirname(__file__)))
_install_dir = setup_path.joinpath("smartsim/_core/_install")

# import buildenv module
buildenv_path = _install_dir.joinpath("buildenv.py")
buildenv_spec = importlib.util.spec_from_file_location("buildenv", str(buildenv_path))
buildenv = importlib.util.module_from_spec(buildenv_spec)
buildenv_spec.loader.exec_module(buildenv)

# import builder module
builder_path = _install_dir.joinpath("builder.py")
builder_spec = importlib.util.spec_from_file_location("builder", str(builder_path))
builder = importlib.util.module_from_spec(builder_spec)
builder_spec.loader.exec_module(builder)

# helper classes for building dependencies that are
# also utilized by the Smart CLI
build_env = buildenv.BuildEnv(checks=False)
versions = buildenv.Versioner()

# check for compatible python versions
if not build_env.is_compatible_python(versions.PYTHON_MIN):
    print("You are using Python {}. Python >={} is required.".format(build_env.python_version,
                                                                     ".".join((versions.PYTHON_MIN))))
    sys.exit(-1)

if build_env.is_windows():
    print("Windows is not supported by SmartSim")
    sys.exit(-1)

# write the SmartSim version into
# smartsim/version.py and to be set as
# __version__ in smartsim/__init__.py
smartsim_version = versions.write_version(setup_path)

class BuildError(Exception):
    pass

# Hacky workaround for solving CI build "purelib" issue
# see https://github.com/google/or-tools/issues/616
class InstallPlatlib(install):
    def finalize_options(self):
        super().finalize_options()
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib

class SmartSimBuild(build_py):

    def run(self):
        database_builder = builder.DatabaseBuilder(build_env(),
                                             build_env.MALLOC,
                                             build_env.JOBS)
        if not database_builder.is_built:
            database_builder.build_from_git(versions.REDIS_URL,
                                         versions.REDIS)

            database_builder.cleanup()

        # run original build_py command
        super().run()


# Tested with wheel v0.29.0
class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name

       We use this because we want to pre-package Redis for certain
       platforms to use.
    """
    def has_ext_modules(_placeholder):
        return True


# Define needed dependencies for the installation
deps = [
    "psutil>=5.7.2",
    "coloredlogs==10.0",
    "tabulate>=0.8.9",
    "redis-py-cluster==2.1.3",
    "redis==3.5.3",
    "tqdm>=4.50.2",
    "filelock>=3.4.2",
    "protobuf~=3.20",
]

# Add SmartRedis at specific version
deps.append("smartredis>={}".format(versions.SMARTREDIS))

extras_require = {
    "dev": [
        "black>=20.8b1",
        "isort>=5.6.4",
        "pylint>=2.6.0",
        "pytest>=6.0.0",
        "pytest-cov>=2.10.1",
        "click==8.0.2",
    ],
    # see smartsim/_core/_install/buildenv.py for more details
    "ml": versions.ml_extras_required(),
}


# rest in setup.cfg
setup(
    version=smartsim_version,
    install_requires=deps,
    packages=["smartsim"],
    package_data={"smartsim": [
        "_core/bin/*",
    ]},
    cmdclass={
        "build_py": SmartSimBuild,
        "install": InstallPlatlib,
    },
    zip_safe=False,
    extras_require=extras_require,
    distclass=BinaryDistribution,
    entry_points={
        "console_scripts": [
            "smart = smartsim._core._cli.__main__:main",
        ]
    }
)
